#*****************************************************************************
#       Copyright (C) 2019 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

from __future__ import print_function

from cysignals.memory cimport check_malloc, sig_free
from libc.string cimport memcpy

def _test_iterator1(int n, int k):
    r"""
    TESTS::

        sage: from surface_dynamics.interval_exchanges.integer_iet import _test_iterator1

        sage: list(_test_iterator1(0, 0))
        [[]]
        sage: list(_test_iterator1(0, 1))
        []
        sage: list(_test_iterator1(1, 0))
        []

        sage: list(_test_iterator1(1, 1))
        [[1L]]
        sage: list(_test_iterator1(2, 1))
        [[2L]]

        sage: list(_test_iterator1(2, 2))
        [[1L, 1L]]
        sage: list(_test_iterator1(3, 2))
        [[2L, 1L], [1L, 2L]]
        sage: list(_test_iterator1(4, 2))
        [[3L, 1L], [2L, 2L], [1L, 3L]]

        sage: list(_test_iterator1(6, 3))
        [[4L, 1L, 1L],
         [3L, 2L, 1L],
         [2L, 3L, 1L],
         [1L, 4L, 1L],
         [3L, 1L, 2L],
         [2L, 2L, 2L],
         [1L, 3L, 2L],
         [2L, 1L, 3L],
         [1L, 2L, 3L],
         [1L, 1L, 4L]]

        sage: list(_test_iterator1(3, 4))
        []
    """
    cdef uint64_t * x

    x = <uint64_t *> check_malloc(k * sizeof(uint64_t))

    if not int_vector_first(x, n, k):
        sig_free(x)
        return
    yield [x[i] for i in range(k)]
    while int_vector_next(x, n, k):
        yield [x[i] for i in range(k)]

    sig_free(x)

def _test_iterator2(int n, int kfree, int ktop, int kbot):
    r"""
    TESTS::

        sage: from surface_dynamics.interval_exchanges.integer_iet import _test_iterator2

        sage: list(_test_iterator2(3, 1, 1, 1))
        [[1L, 1L, 1L]]
        sage: list(_test_iterator2(4, 1, 1, 1))
        [[2L, 1L, 1L]]
        sage: list(_test_iterator2(5, 1, 1, 1))
        [[3L, 1L, 1L], [1L, 2L, 2L]]

        sage: list(_test_iterator2(3, 0, 1, 1))
        []
        sage: list(_test_iterator2(4, 0, 2, 1))
        [[1L, 1L, 2L]]
        sage: list(_test_iterator2(1, 1, 0, 0))
        [[1L]]

        sage: list(_test_iterator2(5, 2, 1, 1))
        [[2L, 1L, 1L, 1L], [1L, 2L, 1L, 1L]]
        sage: list(_test_iterator2(7, 2, 2, 1))
        [[2L, 1L, 1L, 1L, 2L], [1L, 2L, 1L, 1L, 2L]]
        sage: list(_test_iterator2(8, 2, 2, 1))
        [[3L, 1L, 1L, 1L, 2L],
         [2L, 2L, 1L, 1L, 2L],
         [1L, 3L, 1L, 1L, 2L],
         [1L, 1L, 2L, 1L, 3L],
         [1L, 1L, 1L, 2L, 3L]]
        sage: list(_test_iterator2(8, 2, 1, 2))
        [[3L, 1L, 2L, 1L, 1L],
         [2L, 2L, 2L, 1L, 1L],
         [1L, 3L, 2L, 1L, 1L],
         [1L, 1L, 3L, 2L, 1L],
         [1L, 1L, 3L, 1L, 2L]]

         sage: for a in _test_iterator2(5, 1, 1, 1):
         ....:     assert a[0] + 2*a[1] == a[0] + 2*a[2] == 5
    """
    cdef li_vector_iterator_t v

    int_li_vector_init(v, n, kfree, ktop, kbot)

    int_li_vector_prefirst(v)
    while int_li_vector_first_or_next(v):
        yield [(v.x)[i] for i in range(kfree+ktop+kbot)]

    int_li_vector_clear(v)

def perm_to_twin(p):
    r"""
    Return the twin associated to the permutation ``p``.

    EXAMPLES::

        sage: from surface_dynamics.interval_exchanges.integer_iet import perm_to_twin
        sage: perm_to_twin([0,0,1,1,2,2])
        [1, 0, 3, 2, 5, 4]
        sage: perm_to_twin([0,1,2,0])
        Traceback (most recent call last):
        ...
        ValueError: p must contain integer only between 0 and 1
        sage: perm_to_twin([0,0,1,0])
        Traceback (most recent call last):
        ...
        ValueError: each label must appear exactly twice
        sage: perm_to_twin([1,0,1])
        Traceback (most recent call last):
        ...
        ValueError: p must have even length
    """
    n = len(p)
    if n%2 == 1:
        raise ValueError("p must have even length")
    n = n//2
    seen = [None] * len(p)
    t = [None] * len(p)
    for i,j in enumerate(p):
        j = int(j)
        if j < 0 or j >= n:
            raise ValueError("p must contain integer only between 0 and {}".format(n-1))

        ii = seen[j]
        if ii is not None:
            if t[i] is not None or t[ii] is not None:
                raise ValueError("each label must appear exactly twice")
            t[i] = ii
            t[ii] = i
        else:
            seen[j] = i

    return t

cdef set_int_iet(int_iet_t t, top, bot, lengths):
    "Initialize and set t with the data from top, bot, lengths"
    cdef int * labels
    cdef int * twin
    cdef uint64_t * clengths
    cdef int k = len(top)
    cdef int n = (len(top)+len(bot))/2
    cdef int i,j

    # checks and twin construction
    lengths = list(map(int, lengths))
    if len(lengths) != n:
        raise ValueError('invalid lengths')
    p = list(map(int, top + bot))
    if len(p) != 2*n:
        raise ValueError('invalid top, bot')
    python_twin = perm_to_twin(top + bot)

    assert all(l > 0 for l in lengths)
    if sum(lengths[i] for i in top) != sum(lengths[i] for i in bot):
        raise ValueError('different lengths for top and bot')


    # fill C data
    labels = <int *> check_malloc(2*n*sizeof(int));
    twin = <int *> check_malloc(2*n*sizeof(int));
    clengths = <uint64_t *> check_malloc(n*sizeof(uint64_t));

    for i in range(2*n):
        labels[i] = p[i]
        twin[i] = python_twin[i]
    for j in range(n):
        clengths[j] = lengths[j]

    int_iet_init(t, n)
    int_iet_set_labels_and_twin(t, labels, twin, k)
    int_iet_set_lengths(t, clengths)
    sig_free(twin)
    sig_free(labels)
    sig_free(clengths)

    if int_iet_check(t):
        int_iet_clear(t)
        raise RuntimeError("invalid iet")

# we want an iterator over lengths

def _relabel(top, bot):
    r"""
    TESTS::

        sage: from surface_dynamics.interval_exchanges.integer_iet import _relabel
        sage: _relabel([1,0,1],[2,2,0])
        ([1, 0, 1], [2, 2, 0], 1, 1, 1)

        sage: _relabel([1,1,2,2], [0,0])
        ([0, 0, 1, 1], [2, 2], 0, 2, 1)

        sage: _relabel([3,1,3,0,0,4],[4,2,2,1])
        ([2, 0, 2, 3, 3, 1], [1, 4, 4, 0], 2, 2, 1)

        sage: _relabel([3, 2, 0, 1], [1, 2, 0, 3])
        ([0, 1, 2, 3], [3, 1, 2, 0], 4, 0, 0)
    """
    n = (len(top) + len(bot)) / 2
    k = len(top)
    j = 0
    p = top + bot
    twin = perm_to_twin(top + bot)
    for i in range(k):
        if twin[i] >= k:
            p[i] = p[twin[i]] = j
            j += 1
    nfree = j
    for i in range(k):
        if i < twin[i] < k:
            p[i] = p[twin[i]] = j
            j += 1
    ntop = j - nfree
    for i in range(k, 2*n):
        if twin[i] >= k and twin[i] > i:
            p[i] = p[twin[i]] = j
            j += 1
    nbot = j - ntop - nfree
    return p[:k], p[k:], nfree, ntop, nbot

cdef void get_stat(statistics, bint flat, int_iet_t t, uint64_t * widths, uint64_t * heights, int kind):
    r"""
    - ``kind=0``: number of cylinders
    - ``kind=1``: number of components
    - ``kind=2``: tuple of cylinder widths (kind=0 is the length and kind=1 is
      the sum)
    - ``kind=3``: tuple of heights
    - ``kind=4``: tuple of pairs (width, height)
    """
    if kind == 0:
        j = int_iet_num_cylinders(NULL, NULL, t)
        key = j
    elif kind == 1:
        if widths == NULL:
            raise ValueError("widths should be allocated")
        j = int_iet_num_cylinders(widths, NULL, t)
        s = 0
        for i in range(j):
            s += widths[i]
        key = s
    elif kind == 2:
        if widths == NULL:
            raise ValueError("widths should be allocated")
        j = int_iet_num_cylinders(widths, NULL, t)
        key = tuple(sorted(widths[i] for i in range(j)))
    elif kind == 3:
        if heights == NULL:
            raise ValueError("heights should be allocated")
        j = int_iet_num_cylinders(NULL, heights, t)
        key = tuple(sorted(heights[i] for i in range(j)))
    elif kind == 4:
        if widths == NULL or heights == NULL:
            raise ValueError("widths and heights should be allocated")
        j = int_iet_num_cylinders(widths, heights, t)
        key = tuple(sorted((widths[i], heights[i]) for i in range(j)))
    else:
        raise ValueError("unknown kind %d; must be one of 0, 1, 2, 3, 4" % kind)

    if flat:
        (<list> statistics).append(key)
    else:
        if key not in (<dict> statistics):
            (<dict> statistics)[key] = 1
        else:
            (<dict> statistics)[key] += 1

def interval_exchange_statistics(top, bot, uint64_t L, int kind=0, bint flat=False):
    r"""
    Return the statistics about cylinder decomposition of all the integral iet
    with given permutation ``top`` and ``bot`` and total length ``L``.

    An integral interval exchange transformation (or iet for short)
    decomposes into periodic components. We want to study these
    components. Equivalently, such iet corresponds to a permutation
    and we want to study its cycle decomposition.

    INPUT:

    - ``top``, ``bot`` -- top and bottom permutation (as list of
      numbers starting from ``0``)

    - ``L`` (integer) -- the total length of the integral iet to be
      tested

    - ``kind`` -- if ``0`` (default) return statistics about the number of
      cylinders, if ``1`` return statistics about the number of components, if
      ``2`` the keys are tuples of cylinder heights

    OUTPUT: a dictionary whose keys are integers and the value associated to a
    key ``k`` is the number of length data with sum ``L`` which corresponds to
    an iet with data ``top`` and ``bot`` and whose cylinder decomposition has
    ``k`` cylinders.

    EXAMPLES:

    Rotations always give a unique cylinder::

        sage: from surface_dynamics.interval_exchanges.integer_iet import interval_exchange_statistics
        sage: interval_exchange_statistics([0,1], [1,0], 10, 0)
        {1: 9}

     Though, the widths of this unique cylinder is the gcd of the two lengths::

        sage: interval_exchange_statistics([0,1], [1,0], 10, 1)
        {1L: 4, 2L: 4, 5L: 1}
        sage: [gcd(k, 10-k) for k in range(1,10)]
        [1, 2, 1, 2, 5, 2, 1, 2, 1]

    Complete information about the heights of cylinders are obtained by setting ``kind=2``::

        sage: s2 = interval_exchange_statistics([0,1,2,3], [3,2,1,0], 10, 2)
        sage: for k in sorted(s2):
        ....:     print("%-6s: %s" % ('(' + ', '.join('%d' % i for i in k) + ')', s2[k]))
        (1)   : 34
        (1, 1): 16
        (1, 2): 16
        (1, 3): 4
        (1, 4): 6
        (1, 6): 2
        (2, 2): 4
        (2, 3): 2

    You can recover ``kind=0`` and ``kind=1`` by looking and lengths and sums::

        sage: from collections import defaultdict
        sage: s0 = defaultdict(int)
        sage: s1 = defaultdict(int)
        sage: for k, v in s2.items():
        ....:     s0[len(k)] += v
        ....:     s1[sum(k)] += v

        sage: dict(s0) == interval_exchange_statistics([0,1,2,3], [3,2,1,0], 10, 0)
        True
        sage: dict(s1) == interval_exchange_statistics([0,1,2,3], [3,2,1,0], 10, 1)
        True

    To get statistics for the heights, consider using ``kind=3``::

        sage: s3 = interval_exchange_statistics([0,1,2,3], [3,2,1,0], 10, 3)
        sage: for k in sorted(s3):
        ....:     print("%-6s: %s" % ('(' + ', '.join('%d' % i for i in k) + ')', s3[k]))
        (1, 4): 4
        (1, 6): 4
        (1, 8): 4
        (2, 2): 4
        (2, 3): 2
        (2, 4): 8
        (2, 6): 6
        (2, 8): 8
        (3, 4): 2
        (4, 6): 8
        (10)  : 34

    And for both lengths and heights, ``kind=4``::

        sage: s4 = interval_exchange_statistics([0,1,2,3], [3,2,1,0], 10, 4)
        sage: for key in s4:
        ....:     assert sum(l*h for l,h in key) == 10
    """
    cdef int * labels
    cdef int * twin
    cdef uint64_t * widths
    cdef uint64_t * heights
    cdef uint64_t s
    cdef int k = len(top)
    cdef int n = (len(top)+len(bot))/2
    cdef int i
    cdef int_iet_t t
    cdef li_vector_iterator_t v

    if flat:
        statistics = []
    else:
        statistics = {}

    top, bot, kfree, ktop, kbot = _relabel(top, bot)

    p = top + bot
    python_twin = perm_to_twin(p)

    labels = <int *> check_malloc(2 * n * sizeof(int))
    twin = <int *> check_malloc(2 * n * sizeof(int))
    widths = <uint64_t *> check_malloc(n * sizeof(uint64_t))
    heights = <uint64_t *> check_malloc(n * sizeof(uint64_t))

    for i in range(2*n):
        labels[i] = p[i]
        twin[i] = python_twin[i]
    int_iet_init(t, n)

    int_li_vector_init(v, L, kfree, ktop, kbot)
    int_li_vector_prefirst(v)
    while int_li_vector_first_or_next(v):
        int_iet_set_labels_and_twin(t, labels, twin, k)
        int_iet_set_lengths(t, v.x)
        get_stat(statistics, flat, t, widths, heights, kind)

    sig_free(labels)
    sig_free(twin)
    sig_free(widths)
    sig_free(heights)
    int_iet_clear(t)
    int_li_vector_clear(v)

    return statistics

def interval_exchange_statistics_sample(top, bot, uint64_t L, uint64_t sample_size, int kind=0, bint flat=False):
    r"""
    Return the statistics of cylinder decomposition of a random sample of ``sample_size``
    integral iet with given permutation ``top`` and ``bot``.

    INPUT:

    - ``top``, ``bot`` -- top and bottom permutation (as list of
      numbers starting from ``0``)

    - ``L`` (integer) -- a magnitude order for the total length of the integral
      iet to be generated

    - ``kind`` -- if ``0`` (default) return statistics about the number of
      cylinders, if ``1`` return statistics about the number of components, if
      ``2`` the keys are tuples of cylinder heights

    OUTPUT: a dictionary whose keys are integers and the value associated to a
    key ``k`` is the number of length data with sum ``L`` which corresponds to
    an iet with data ``top`` and ``bot`` and whose cylinder decomposition has
    ``k`` cylinders.

    EXAMPLES:

    Rotations always give a unique cylinder::

        sage: from surface_dynamics.interval_exchanges.integer_iet import interval_exchange_statistics
        sage: interval_exchange_statistics([0,1],[1,0],10,0)
        {1: 9}

     Though, the widths of this unique cylinder is the gcd of the two lengths::

        sage: interval_exchange_statistics([0,1],[1,0],10,1)
        {1L: 4, 2L: 4, 5L: 1}
        sage: [gcd(k, 10-k) for k in range(1,10)]
        [1, 2, 1, 2, 5, 2, 1, 2, 1]

    Complete information about the heights of cylinders are obtained by setting ``kind=2``::

        sage: s2 = interval_exchange_statistics([0,1,2,3], [3,2,1,0], 10, 2)
        sage: for k in sorted(s2):
        ....:     print("%-6s: %s" % ('(' + ', '.join('%d' % i for i in k) + ')', s2[k]))
        (1)   : 34
        (1, 1): 16
        (1, 2): 16
        (1, 3): 4
        (1, 4): 6
        (1, 6): 2
        (2, 2): 4
        (2, 3): 2

    You can recover ``kind=0`` and ``kind=1`` by looking and lengths and sums::

        sage: from collections import defaultdict
        sage: s0 = defaultdict(int)
        sage: s1 = defaultdict(int)
        sage: for k, v in s2.items():
        ....:     s0[len(k)] += v
        ....:     s1[sum(k)] += v

        sage: dict(s0) == interval_exchange_statistics([0,1,2,3], [3,2,1,0], 10, 0)
        True
        sage: dict(s1) == interval_exchange_statistics([0,1,2,3], [3,2,1,0], 10, 1)
        True
    """
    cdef int * labels
    cdef int * twin
    cdef uint64_t * widths
    cdef uint64_t * heights
    cdef uint64_t s
    cdef int k = len(top)
    cdef int n = (len(top)+len(bot))/2
    cdef int i
    cdef int count
    cdef int_iet_t t

    if flat:
        statistics = []
    else:
        statistics = {}

    top, bot, kfree, ktop, kbot = _relabel(top, bot)

    p = top + bot
    python_twin = perm_to_twin(p)

    labels = <int *> check_malloc(2 * n * sizeof(int))
    twin = <int *> check_malloc(2 * n * sizeof(int))
    widths = <uint64_t *> check_malloc(n * sizeof(uint64_t))
    heights = <uint64_t *> check_malloc(n * sizeof(uint64_t))

    for i in range(2*n):
        labels[i] = p[i]
        twin[i] = python_twin[i]
    int_iet_init(t, n)

    for count in range(sample_size):
        int_iet_set_labels_and_twin(t, labels, twin, k)
        int_iet_set_random_lengths(t, L)
        get_stat(statistics, flat, t, widths, heights, kind)

    sig_free(labels)
    sig_free(twin)
    sig_free(widths)
    sig_free(heights)
    int_iet_clear(t)

    return statistics


# there are several interesting statistics one might want to compute
# such as
#  - number of cylinders
#  - sum of heights of cylinders
#  - mean ratio in view of Siegel Veech constants (not available for now) as
#    we would need the circumferences of cylinders as well
def cylinder_statistics(top, bot, uint64_t L, int kind=0, bint flat=False):
    r"""
    Return the statistics of the number of cylinders for a given total length

    INPUT:

    - ``top`` -- (list) composition of the top of the cylinder

    - ``bot`` -- (list) composition of the bottom of the cylinder
    
    - ``L`` -- (positive integer) the length we are considering

    - ``kind`` -- kind of statistics to gather
      - ``0`` number of cylinders
      - ``1`` number of components
      - ``2`` tuple of widths
      - ``3`` tuple of heights
      - ``4`` tuple of pairs (width, height)

    EXAMPLES::

        sage: from surface_dynamics.interval_exchanges.integer_iet import cylinder_statistics

    The case of H(0,0)::

        sage: top = [0, 1]
        sage: bot = [1, 0]
        sage: for n in range(2,20):
        ....:     s = cylinder_statistics(top, bot, n)
        ....:     print("%2d : %3d %3d %3d" %(n, s[1], s[2], s[1] + s[2]))
         2 :   1   1   2
         3 :   4   2   6
         4 :   7   5  12
         5 :  16   4  20
         6 :  15  15  30
         7 :  36   6  42
         8 :  35  21  56
         9 :  52  20  72
        10 :  53  37  90
        11 : 100  10 110
        12 :  65  67 132
        13 : 144  12 156
        14 : 115  67 182
        15 : 132  78 210
        16 : 155  85 240
        17 : 256  16 272
        18 : 165 141 306
        19 : 324  18 342

    Which are the same numbers as Q(0,-1^4)::

        sage: top1 = [0, 1]
        sage: bot1 = [1, 0]
        sage: top2 = [0, 0]
        sage: bot2 = [1, 1, 2, 2]
        sage: for _ in range(20):
        ....:     n = randint(2, 100)
        ....:     s1 = cylinder_statistics(top1, bot1, n)
        ....:     s2 = cylinder_statistics(top2, bot2, 2*n)
        ....:     assert sorted(s1.keys()) == sorted(s2.keys())
        ....:     assert all(2*s1[i] == s2[i] for i in s1)
    """
    cdef int * labels1
    cdef int * labels2
    cdef int * twin1
    cdef int * twin2
    cdef uint64_t * clengths
    cdef int k1 = len(top)
    cdef int k2 = k1 + 1
    cdef int n = (len(top)+len(bot))/2
    cdef int i,j
    cdef uint64_t twist
    cdef uint64_t * widths
    cdef uint64_t * heights
    cdef int_iet_t t1, t2
    cdef li_vector_iterator_t v

    if flat:
        statistics = []
    else:
        statistics = {}

    top, bot, kfree, ktop, kbot = _relabel(top, bot)

    p1 = top + bot
    p2 = [n] + p1 + [n]
    python_twin1 = perm_to_twin(p1)
    python_twin2 = perm_to_twin(p2)

    labels1 = <int *> check_malloc(2*n*sizeof(int))
    twin1 = <int *> check_malloc(2*n*sizeof(int))
    labels2 = <int *> check_malloc(2*(n+1)*sizeof(int))
    twin2 = <int *> check_malloc(2*(n+1)*sizeof(int))
    widths = <uint64_t *> check_malloc((n+1) * sizeof(uint64_t))
    heights = <uint64_t *> check_malloc((n+1) * sizeof(uint64_t))

    clengths = <uint64_t *> check_malloc((n+1)*sizeof(uint64_t))

    for i in range(2*n):
        labels1[i] = p1[i]
        twin1[i] = python_twin1[i]
    int_iet_init(t1, n)

    for i in range(2*(n+1)):
        labels2[i] = p2[i]
        twin2[i] = python_twin2[i]
    int_iet_init(t2, n+1)

    int_li_vector_init(v, L, kfree, ktop, kbot)
    int_li_vector_prefirst(v)
    while int_li_vector_first_or_next(v):
        memcpy(clengths, v.x, n * sizeof(uint64_t))

        # twist 0 case
        int_iet_set_labels_and_twin(t1, labels1, twin1, k1)
        int_iet_set_lengths(t1, clengths)
        get_stat(statistics, flat, t1, widths, heights, kind)

        # positive twists case
        for twist in range(1, L):
            clengths[n] = twist
            int_iet_set_labels_and_twin(t2, labels2, twin2, k2)
            int_iet_set_lengths(t2, clengths)
            get_stat(statistics, flat, t2, widths, heights, kind)

    sig_free(labels1)
    sig_free(twin1)
    sig_free(labels2)
    sig_free(twin2)
    sig_free(clengths)
    sig_free(widths)
    sig_free(heights)
    int_iet_clear(t1)
    int_iet_clear(t2)
    int_li_vector_clear(v)

    return statistics

def cylinder_number(top, bot, lengths):
    r"""
    Return the number of cylinders and of a given interval exchange

    EXAMPLES::

        sage: from surface_dynamics.interval_exchanges.integer_iet import cylinder_number

        sage: top = [0, 1]
        sage: bot = [1, 0]
        sage: cylinder_number(top, bot, [13, 24])
        1

        sage: top = [0, 0]
        sage: bot = [1, 1, 2, 2, 3, 3, 4, 4]
        sage: cylinder_number(top, bot, [10, 4, 3, 2, 1])
        2

        sage: top = [0, 0, 1, 1, 2, 2]
        sage: bot = [3, 3, 4, 4, 5, 5, 6, 6]
        sage: cylinder_number(top, bot, [8, 7, 10, 8, 10, 1, 6])
        3
    """
    cdef int_iet_t t
    set_int_iet(t, top, bot, lengths)
    cdef int res = int_iet_num_cylinders(NULL, NULL, t)
    int_iet_clear(t)
    return res

def cylinder_widths_and_heights(top, bot, lengths):
    r"""
    Return the list of pairs ``(width, height)`` of each periodic component of the
    iet determined by ``top``, ``bot`` and ``lengths``.

    EXAMPLES::

        sage: from surface_dynamics.interval_exchanges.integer_iet import cylinder_widths_and_heights

        sage: cylinder_widths_and_heights([0,1],[1,0],[5,2])
        [(1L, 7L)]
        sage: cylinder_widths_and_heights([0,1],[1,0],[8,12])
        [(4L, 5L)]

        sage: cylinder_widths_and_heights([0,1,2],[2,1,0],[5,3,2])
        [(1L, 10L)]

        sage: cylinder_widths_and_heights(list(range(8)), list(range(7,-1,-1)), [1386,924,660,495,385,308,252,210])
        [(2L, 930L), (3L, 920L)]
    """
    cdef int_iet_t t
    cdef uint64_t * widths
    cdef uint64_t * heights
    set_int_iet(t, top, bot, lengths)
    widths = <uint64_t *> check_malloc((len(bot) + len(top)) / 2 * sizeof(uint64_t))
    heights = <uint64_t *> check_malloc((len(bot) + len(top)) / 2 * sizeof(uint64_t))
    cdef int res = int_iet_num_cylinders(widths, heights, t)
    int_iet_clear(t)
    output =  [(widths[i], heights[i]) for i in range(res)]
    sig_free(widths)
    sig_free(heights)
    output.sort()
    return output

def cylinder_widths(top, bot, lengths):
    r"""
    Return the widths of cylinders of the interval exchange determined by ``top``, ``bot``, ``lengths``.

    EXAMPLES::

        sage: from surface_dynamics.interval_exchanges.integer_iet import cylinder_widths

        sage: top = [0, 1]
        sage: bot = [1, 0]
        sage: cylinder_widths(top, bot, [13, 24])
        [1L]

        sage: top = [0, 0]
        sage: bot = [1, 1, 2, 2, 3, 3, 4, 4]
        sage: cylinder_widths(top, bot, [10, 4, 3, 2, 1])
        [1L, 2L]

    Genus 0 examples::

        sage: top = [0, 0, 1, 1, 2, 2]
        sage: bot = [3, 3, 4, 4, 5, 5, 6, 6]
        sage: cylinder_widths(top, bot, [8, 7, 10, 8, 10, 1, 6])
        [1L, 2L, 8L]
        sage: cylinder_widths(top, bot, [1298, 814, 23, 140, 1034, 819, 142])
        [2L, 3L]
        sage: cylinder_widths(top, bot, [78745, 39773, 19984, 21665, 31239, 437, 85161])
        [1L, 1L, 1L]
        sage: cylinder_widths(top, bot, [57453, 9483, 110211, 100249, 36644, 9853, 30401])
        [1L, 1L, 3L]

        sage: top = [0, 0, 1, 1]
        sage: bot = [2, 2, 3, 3, 4, 4, 5, 5]
        sage: cylinder_widths(top, bot, [746825, 30952, 206252, 348, 538642, 32535])
        [1L]
        sage: cylinder_widths(top, bot, [219989, 91122, 70542, 187806, 5742, 47021])
        [1L, 2L]
        sage: cylinder_widths(top, bot, [49622, 73834, 4390, 46944, 43358, 28764])
        [2L, 10L]
        sage: cylinder_widths(top, bot, [25489, 109646, 25473, 7492, 11490, 90680])
        [1L, 2L]

    Genus 3 example::

        sage: top = [0, 1, 2, 3, 4, 5, 3, 4, 6]
        sage: bot = [7, 8, 1, 7, 8, 2, 5, 6, 0]
        sage: cylinder_widths(top, bot, [8, 10, 8, 12, 7, 3, 10, 11, 8])
        [1L]
        sage: cylinder_widths(top, bot, [9, 10, 8, 1, 6, 3, 5, 4, 3])
        [1L, 1L]
        sage: cylinder_widths(top, bot, [11, 10, 8, 12, 7, 3, 10, 11, 8])
        [1L, 1L, 2L]
        sage: cylinder_widths(top, bot, [948, 803, 775, 12, 7, 951, 10, 11, 8])
        [1L]
    """
    cdef int_iet_t t
    cdef uint64_t * widths
    set_int_iet(t, top, bot, lengths)
    widths = <uint64_t *> check_malloc((len(bot) + len(top)) / 2 * sizeof(uint64_t))
    cdef int res = int_iet_num_cylinders(widths, NULL, t)
    int_iet_clear(t)
    output =  [widths[i] for i in range(res)]
    sig_free(widths)
    output.sort()
    return output



r"""
Some extra ppl utilities

ppl is a library to deal with rational polytopes. pplpy is a thin
Python wrapper. This file provides some helper functions to use it.
"""
#*****************************************************************************
#       Copyright (C) 2019 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

import ppl

def ppl_check_non_negative_cone(C):
    r"""
    Check whether the input cone ``C`` is a subcone of the positive cone

    EXAMPLES::

        sage: from surface_dynamics.misc.ppl_utils import ppl_check_non_negative_cone # optional - pplpy
        sage: import ppl    # optional - pplpy

        sage: gs = ppl.Generator_System()     # optional - pplpy
        sage: gs.insert(ppl.point())          # optional - pplpy
        sage: gs.insert(ppl.ray(ppl.Variable(0) + ppl.Variable(2))) # optional - pplpy
        sage: ppl_check_non_negative_cone(ppl.C_Polyhedron(gs)) # optional - pplpy

        sage: gs.insert(ppl.ray(-ppl.Variable(0) - ppl.Variable(2))) # optional - pplpy
        sage: ppl_check_non_negative_cone(ppl.C_Polyhedron(gs)) # optional - pplpy
        Traceback (most recent call last):
        ...
        ValueError: C must be a subcone of the non-negative cone

        sage: gs = ppl.Generator_System() # optional - pplpy
        sage: gs.insert(ppl.point(ppl.Variable(2))) # optional - pplpy
        sage: ppl_check_non_negative_cone(ppl.C_Polyhedron(gs)) # optional - pplpy
        Traceback (most recent call last):
        ...
        ValueError: the cone does not contain zero
        sage: gs.insert(ppl.point()) # optional - pplpy
        sage: ppl_check_non_negative_cone(ppl.C_Polyhedron(gs)) # optional - pplpy
        Traceback (most recent call last):
        ...
        ValueError: should have only zero as vertices
    """
    if not isinstance(C, ppl.C_Polyhedron):
        raise ValueError('C must be a polyhedron in the right ambient space')

    n = C.space_dimension()
    if not C.contains(ppl.C_Polyhedron(ppl_zero_point(n))):
        raise ValueError('the cone does not contain zero')
    if not ppl_positive_cone(n).contains(C):
        raise ValueError('C must be a subcone of the non-negative cone')
    for g in C.generators():
        if g.is_point() and not g.is_equivalent_to(ppl_zero_point(n)):
            raise ValueError('should have only zero as vertices'.format(g))
        if g.is_line():
            raise ValueError('the cone contains a line')

def ppl_zero_point(n):
    r"""
    Return the origin in R^n

    EXAMPLES::

        sage: from surface_dynamics.misc.ppl_utils import ppl_zero_point # optional - pplpy
        sage: ppl_zero_point(3)   # optional - pplpy
        point(0/1, 0/1, 0/1)
        sage: ppl_zero_point(5)   # optional - pplpy
        point(0/1, 0/1, 0/1, 0/1, 0/1)
    """
    z = ppl.point()
    z.set_space_dimension(n)
    return z

def ppl_positive_cone(n):
    r"""
    Return the positive cone in R^n

    EXAMPLES::

        sage: from surface_dynamics.misc.ppl_utils import ppl_positive_cone # optional - pplpy
        sage: C = ppl_positive_cone(3)  # optional - pplpy
        sage: C.minimized_generators()  # optional - pplpy
        Generator_System {point(0/1, 0/1, 0/1), ray(0, 0, 1), ray(0, 1, 0), ray(1, 0, 0)}
    """
    gs = ppl.Generator_System(ppl_zero_point(n))
    l = [0]*n
    for i in range(n):
        gs.insert(ppl.ray(ppl.Variable(i)))
    return ppl.C_Polyhedron(gs)

def ppl_cone(rays):
    r"""
    Convert the list ``rays`` into a ppl cone

    EXAMPLES::

        sage: from surface_dynamics.misc.ppl_utils import ppl_cone  # optional - pplpy
        sage: C = ppl_cone([(0,1,2),(1,1,1),(1,0,1)])   # optional - pplpy
        sage: C.minimized_generators()   # optional - pplpy
        Generator_System {point(0/1, 0/1, 0/1), ray(0, 1, 2), ray(1, 0, 1), ray(1, 1, 1)}
    """
    n = len(rays[0])
    gs = ppl.Generator_System(ppl_zero_point(n))
    for r in rays:
        gs.insert(ppl.ray(sum(int(j) * ppl.Variable(i) for i,j in enumerate(r))))
    return ppl.C_Polyhedron(gs)


def ppl_convert(P):
    r"""
    Convert a Sage polyhedron to a ppl polyhedron

    EXAMPLES::

        sage: from surface_dynamics.misc.ppl_utils import ppl_convert  # optional - pplpy
        sage: P = ppl_convert(Polyhedron(vertices=[(0,1,0),(1,0,1)], rays=[(0,0,1),[3,2,1]]))  # optional - pplpy
        sage: P.minimized_generators()  # optional - pplpy
        Generator_System {ray(0, 0, 1), point(0/1, 1/1, 0/1), point(1/1, 0/1, 1/1), ray(3, 2, 1)}
    """
    if isinstance(P, ppl.C_Polyhedron):
        return P
    gs = ppl.Generator_System()
    for v in P.vertices_list():
        gs.insert(ppl.point(sum(int(j) * ppl.Variable(i) for i,j in enumerate(v))))
    for r in P.rays_list():
        gs.insert(ppl.ray(sum(int(j) * ppl.Variable(i) for i,j in enumerate(r))))
    for l in P.lines_list():
        gs.insert(ppl.line(sum(int(j) * ppl.Variable(i) for i,j in enumerate(l))))
    return ppl.C_Polyhedron(gs)




r"""
Simplicial complex, homology of surfaces and translation surfaces

In this module are implemented simple homology computation for translation
surfaces. There are three main classes:

- :class:`RibbonGraph`: decomposition of a surface into polygons. The
  combinatorics is stored as a triple of permutations `v` (vertices), `e`
  (edges), `f` (faces) so that the product `vef` is the identity. The domain of
  the permutations correspond to the half edges or *darts*. The permutation `e`
  is an involution so that `e(i)` is the other half of the edge starting at
  `i`. The fixed points of `e` corresponds to edge glued to themselves. The
  permutation `v` is obtained by turning around a vertex, while `f` turning
  around a face.

- :class:`RibbonGraphWithAngles`: a ribbon graph with an additional angle
  structure.

- :class:`RibbonGraphWithHolonomies`: a ribbon graph with an additional holonomy
  structure on its edges.

EXAMPLES::

    sage: from surface_dynamics import *

To create a ribbon graph you just need to fix two of the permutations `v`, `e`,
`f`::

    sage: R = RibbonGraph(vertices='(0,1,4,3)(5,2)',edges='(0,3)(1,2)(4,5)')
    sage: R
    Ribbon graph with 2 vertices, 3 edges and 3 faces

The vertices, edges and faces are by definition the cycles of the permutation.
Calling the method :meth:`~RibbonGraph.vertices`, :meth:`~RibbonGraph.edges` or
:meth:`~RibbonGraph.faces` gives you access to these cycles::

    sage: R.vertices()
    [[0, 1, 4, 3], [2, 5]]
    sage: R.edges()
    [[0, 3], [1, 2], [4, 5]]
    sage: R.faces()
    [[0, 4, 2], [1, 5], [3]]

Given a half edge (i.e. a dart), you can get the index of the vertex, edge or
face it belongs with the methods :meth:`~RibbonGraph.dart_to_vertex`,
:meth:`~RibbonGraph.dart_to_edge` and :meth:`~RibbonGraph.dart_to_edge`::

    sage: R.dart_to_vertex(1)
    0
    sage: 1 in R.vertices()[0]
    True
    sage: R.dart_to_vertex(2)
    1
    sage: 2 in R.vertices()[1]
    True

    sage: R.dart_to_edge(3)
    0
    sage: R.dart_to_edge(4)
    2

    sage: R.dart_to_face(4)
    0
    sage: R.dart_to_face(3)
    2

To initialize a ribbon graph with angles, you have to input the standard data to
initialize a ribbon graph plus a list of positive rational numbers which
corresponds to the angles between darts (more precisely, the number at position
i is the angle between i and v(i))::

    sage: e = '(0,1)(2,3)'
    sage: f = '(0,2,1,3)'
    sage: a = [1/2,1/2,1/2,1/2]
    sage: r = RibbonGraphWithAngles(edges=e,faces=f,angles=a)
    sage: r.spin_parity()
    1

TODO:

- Oriented ribbon graphs are *exactly* separatrix diagrams. An orientation on a Ribbon
  graphs is a choice of orientation for each vertex so that each face has all of its
  edges oriented in the same direction as we go along the boundaries.
"""
#*****************************************************************************
#       Copyright (C) 2019 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

from __future__ import print_function, absolute_import
from six.moves import range, map, filter, zip

from surface_dynamics.misc.permutation import (perm_init, constellation_init, perm_cycles, perm_invert,
        perm_check, perm_compose, equalize_perms, perm_orbit)

from sage.misc.cachefunc import cached_method

from sage.modules.free_module import FreeModule
from sage.matrix.constructor import matrix, identity_matrix

from sage.structure.sage_object import SageObject
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ


# TODO: introduce an oriented ribbon graph class
# an oriented ribbon graph is a ribbon graph such that one can choose
# coherently an orientation of each edge so that each face gets
# all its edges oriented the same way
# (the dual corresponds in having all vertices being either sources or sinks)
# we encode this by choosing by labeling for each edge with the smaller edge index

class RibbonGraph(SageObject):
    r"""
    Generic class for Ribbon graph.

    A Ribbon graph (or fat graph or combinatorial map) is a graph embedded in a
    surface. This class uses representation as a triple ``(v,e,f)`` of
    permutations such that `vef = 1` and the action of the group generated by
    `v,e,f` acts transitvely in the the domain. The cycles of ``v`` are
    considered as vertices, the ones of ``e`` are considered as edges and the
    ones of ``f`` as the faces. Each element of the domain is a half-edge which
    is called a *dart*. A dart is also associated to an oriented edge.

    The domain of the permutations must be a subset of [0, ..., N-1] for some N.
    The edges are always considered to be (i, ~i) where i is the bit complement
    of the integer i (~0 = -1, ~-3 = 2). So that an edge always has a canonical
    representative number given by the non-negative version.

    A dense ribbon graph has the following attributes

      - total_darts - non negative integer - the total number darts
      - num_darts - non negative integer - the number of active darts
      - active_darts - bitset - list of lengths _total_darts with True or
        False. The position i is True if i is an active dart.

      - vertices, vertices_inv - list - partial permutations of [0,N] which are
        inverse of each other
      - vertex_cycles - the cycles of the partial permutation vertices
      - dart_to_vertex_index

      - edges, edges_inv - list - partial permutations of [0,N] which are
        inverse of each other
      - edge_cycles - the cycles of the partial permutation edge
      - dart_to_edge_index

      - faces, faces_inv - list - partial permutations of [0,N] which are
        inverse of each other
      - face_cycles - the cycles of the partial permutation faces
      - dart_to_face_index

    EXAMPLES::

        sage: from surface_dynamics import *

        sage: RibbonGraph([],[],[])
        Ribbon graph with 1 vertex, 0 edge and 1 face
        sage: RibbonGraph('()','(0,1)','(0,1)')
        Ribbon graph with 2 vertices, 1 edge and 1 face

        sage: G = RibbonGraph('(0,3)(1,2)','(0,1)(2,3)','(0,2)(1,3)')
        sage: G
        Ribbon graph with 2 vertices, 2 edges and 2 faces
        sage: G.darts()
        [0, 1, 2, 3]
        sage: G.genus()
        0

        sage: G = RibbonGraph(edges='(0,2)(1,3)(4,6)(5,7)',faces='(0,1,2,3,4,5,6,7)')
        sage: G
        Ribbon graph with 1 vertex, 4 edges and 1 face
        sage: G.darts()
        [0, 1, 2, 3, 4, 5, 6, 7]
        sage: G.genus()
        2

        sage: G = RibbonGraph(vertices='(0,2,3,6)(1,4,5,7)')
        sage: G
        Ribbon graph with 2 vertices, 4 edges and 4 faces
        sage: G.edges()
        [[0, 1], [2, 3], [4, 5], [6, 7]]
        sage: G.genus()
        0
    """
    def __init__(self, vertices=None, edges=None, faces=None, connected=True, check=True):
        r"""
        TESTS::

            sage: from surface_dynamics import RibbonGraph
            sage: RibbonGraph([0r,1r],[1r,0r],[1r,0r])
            Ribbon graph with 2 vertices, 1 edge and 1 face

            sage: RibbonGraph([-1,1,-1,3],[-1,3,-1,1],[-1,3,-1,1])
            Ribbon graph with 2 vertices, 1 edge and 1 face

            sage: RibbonGraph(vertices='(1)(5)', edges='(1,5)', faces='(1,5)')
            Ribbon graph with 2 vertices, 1 edge and 1 face
        """
        vertices, edges, faces = constellation_init(vertices, edges, faces, check=check)

        self._total_darts = total_darts = len(vertices)
        num_darts = 0
        self._active_darts = [True] * total_darts
        for i,(jv, je, jf) in enumerate(zip(vertices, edges, faces)):
            if jv == -1:
                if je != -1 or jf != -1:
                    raise ValueError("inconsistent permutations")
                self._active_darts[i] = False
            else:
                if je == -1 or jf == -1:
                    raise ValueError("inconsistent permutations")
                num_darts += 1

        self._num_darts = num_darts

        self._vertices = vertices
        self._vertex_cycles = perm_cycles(vertices, singletons=True)
        self._dart_to_vertex_index = [None] * total_darts
        for i,c in enumerate(self._vertex_cycles):
            for j in c:
                self._dart_to_vertex_index[j] = i

        self._edges = edges
        self._edge_cycles = perm_cycles(edges, singletons=True)
        self._dart_to_edge_index = [None] * total_darts
        for i,edge in enumerate(self._edge_cycles):
            for e in edge:
                self._dart_to_edge_index[e] = i

        self._faces = faces
        self._face_cycles = perm_cycles(faces, singletons=True)
        self._dart_to_face_index = [None] * total_darts
        for i,c in enumerate(self._face_cycles):
            for j in c:
                self._dart_to_face_index[j] = i

        self._vertices_inv = [None] * total_darts
        self._edges_inv = [None] * total_darts
        self._faces_inv = [None] * total_darts
        for i in range(total_darts):
            if self._vertices[i] != -1:
                self._vertices_inv[self._vertices[i]] = i
                self._edges_inv[self._edges[i]] = i
                self._faces_inv[self._faces[i]] = i

        self._connected = bool(connected)
        if check:
            self._check()

    def _repr_(self):
        r"""
        String representation.

        TESTS::

            sage: from surface_dynamics import *

            sage: RibbonGraph(edges='(0,1)(2,3)',faces='(0,1,2,3)')._repr_()
            'Ribbon graph with 3 vertices, 2 edges and 1 face'
            sage: RibbonGraph(edges='(0,1)',faces='(0)(1)')._repr_()
            'Ribbon graph with 1 vertex, 1 edge and 2 faces'
            sage: RibbonGraph(edges='(0,1)(2,3)',faces='(0,2)(1,3)')._repr_()
            'Ribbon graph with 2 vertices, 2 edges and 2 faces'
        """
        n = self.num_vertices()
        if n <= 1:
            vert_str = "%d vertex" %n
        else:
            vert_str = "%d vertices" %n
        n = self.num_edges()
        if n <= 1:
            edge_str = "%d edge" %n
        else:
            edge_str = "%d edges" %n
        n = self.num_faces()
        if n <= 1:
            face_str = "%d face" %n
        else:
            face_str = "%d faces" %n
        return "Ribbon graph with %s, %s and %s" %(vert_str,edge_str,face_str)

    @cached_method
    def _symmetric_group(self):
        from sage.groups.perm_gps.permgroup_named import SymmetricGroup
        return SymmetricGroup([i for i,j in enumerate(self._active_darts) if j])

    def monodromy_group(self):
        r"""
        Return the group generated by the three defining permutations.

        EXAMPLES::

            sage: from surface_dynamics import RibbonGraph

            sage: r1 = RibbonGraph(vertices='(0)(2)(1,3,4,5)', edges='(0,1)(2,3)(4,5)')
            sage: G1 = r1.monodromy_group()
            sage: G1
            Subgroup ...
            sage: G1.is_isomorphic(SymmetricGroup(5))
            True

            sage: r2 = RibbonGraph(vertices='(0)(2)(1,3,4)(5,6,7)', edges='(0,1)(2,3)(4,5)(6,7)')
            sage: G2 = r2.monodromy_group()
            sage: G2
            Subgroup ...
            sage: G2.is_isomorphic(PSL(2,7))
            True

            sage: r3 = RibbonGraph(vertices='(1)(5)', edges='(1,5)', faces='(1,5)')
            sage: r3.monodromy_group()
            Subgroup ...
        """
        S = self._symmetric_group()
        v = S([i for i in self._vertices if i != -1])
        e = S([i for i in self._edges if i != -1])
        f = S([i for i in self._faces if i != -1])
        assert (v*e*f).is_one()
        return S.subgroup([v,e,f])

    def automorphism_group(self, fix_vertices=False, fix_edges=False, fix_faces=False):
        r"""
        Return the automorphism group of this ribbon graph.

        The automorphism is simply the intersection of the centralizers
        of the defining permutations.

        EXAMPLES::

            sage: from surface_dynamics import *
            sage: r = RibbonGraph('(1,6,5)(2,3,4)', '(1,2)(3,4)(5,6)', '(1,4,2,5)(3)(6)')
            sage: r.automorphism_group().cardinality()
            2
            sage: r.automorphism_group(fix_faces=True).cardinality()
            1

            sage: r = RibbonGraph('(1,4,5)(2,6,3)', '(1,2)(3,4)(5,6)', '(1,3)(2,5)(4,6)')
            sage: r.automorphism_group().cardinality()
            6

        Examples in genus 1::

            sage: r = RibbonGraph('(1,5,4)(6,3,2)', '(1,2)(3,4)(5,6)', '(1,3,5,2,4,6)')
            sage: A = r.automorphism_group()
            sage: A
            Subgroup ...
            sage: A.cardinality()
            6
            sage: r.automorphism_group(fix_faces=True) == A
            True

            sage: r = RibbonGraph('(1,3,2,4)', '(1,2)(3,4)', '(1,3,2,4)')
            sage: r.automorphism_group().cardinality()
            4
        """
        C = self._symmetric_group().centralizer(self.monodromy_group())
        if C.cardinality().is_one():
            return C
        
        if fix_vertices or fix_edges or fix_faces:
            S = self._symmetric_group()
            B = []
            if fix_vertices:
                B.append(self._vertex_cycles)
            if fix_edges:
                B.append(self._edge_cycles)
            if fix_faces:
                B.append(self._face_cycles)
            for BB in B:
                gens = []
                for s in BB:
                    if len(s) == 1:
                        continue
                    gens.append(S([(s[0], s[1])]))
                    if len(s) > 2:
                        gens.append(S([tuple(s)]))
                C = C.intersection(S.subgroup(gens))

        return C


    # TODO: perform the expansion as quasi-polynomial
    # TODO: see how the recursion formula translates on this generating series
    def length_rational_fraction(self, var='b'):
        r"""
        Return the generating series for the number of lengths with the given boundaries
        """
        from sage.symbolic.ring import SR

        F = SR.one()

        for dart in range(self._total_darts):
            if not self._active_darts[dart]:
                continue
            i = self._dart_to_edge_index[dart]
            j1, j2 = self._edge_cycles[i]
            if j1 == dart:
                continue
            else:
                assert j2 == dart

            f1 = self._dart_to_face_index[j1]
            f2 = self._dart_to_face_index[j2]

            b1 = SR.var('%s%d' %(var, f1))
            b2 = SR.var('%s%d' %(var, f2))

            F *= b1*b2 / (1 - b1*b2)

        return F


    def add_extra_darts(self, n):
        r"""
        Add extra darts to support a total of ``2 n`` darts.
        """
        m = self._total_darts
        if 2*n > m:
            self._total_darts = int(2*n)
            for p in [self._vertices, self._vertices_inv,
                      self._edges, self._edges_inv,
                      self._faces, self._faces_inv]:
                p.extend([None] * (2*n - m))
            self._active_darts.extend([False] * (2*n - m))
            self._check()

    def _check(self):
        r"""
        Check that the data of the Ribbon graph is coherent
        """
        if len(self._active_darts) != self._total_darts:
            raise ValueError("the length of active darts is not total_darts")
        if self._active_darts.count(True) != self._num_darts:
            raise ValueError("the number of darts do not coincide with active darts")

        for i in range(self._total_darts):
            if self._active_darts[i]:
                if self._vertices[i] == -1 or self._vertices_inv[i] is None:
                    raise ValueError("dart %d is active but has no vertex" %i)
                if self._edges[i] == -1 or self._edges_inv[i] is None:
                    raise ValueError("dart %d is active but has no edge" %i)
                if self._faces[i] == -1 or self._faces_inv[i] is None:
                    raise ValueError("dart %d is active but has no face" %i)

                if self._vertices[self._vertices_inv[i]] != i:
                    raise ValueError("vertices is not the inverse of vertices_inv")
                if self._edges[self._edges_inv[i]] != i:
                    raise ValueError("edges is not the inverse of edges_inv")
                if self._faces[self._faces_inv[i]] != i:
                    raise ValueError("faces is not the inverse of faces_inv")

                if self._faces[self._edges[self._vertices[i]]] != i:
                    raise ValueError("the Ribbon graph condition vef=() is not satisfied for %d" %i)

            else:
                if self._vertices[i] != -1 or self._vertices_inv[i] is not None:
                    raise ValueError("dart %d is not active but has a vertex" %i)
                if self._edges[i] != -1 or self._edges_inv[i] is not None:
                    raise ValueError("dart %d is not active but has an edge" %i)
                if self._faces[i] != -1 or self._faces_inv[i] is not None:
                    raise ValueError("dart %d is not active but has a face" %i)

        if self._connected and not self.is_connected(force_computation=True):
            raise ValueError("the graph is not connected")

    def is_connected(self, force_computation=False):
        if not force_computation and self._connected:
            return True

        from sage.graphs.graph import Graph
        G = Graph(loops=True, multiedges=True)
        for i in range(self._total_darts):
            if self._active_darts[i]:
                G.add_edge(i,self._vertices[i])
                G.add_edge(i,self._edges[i])
                G.add_edge(i,self._faces[i])
        return G.is_connected()

    def relabel(self, perm=None):
        r"""
        perm is a of range(0,N)

        If ``perm`` is None, relabel the darts on 0,2M keeping the relative
        order of the darts.
        """
        if perm is None:
            perm=[None]*self.num_darts()
            k = 0
            for i in range(self.num_darts()):
                if self._active_darts[i]:
                    perm[i] = k
                    k += 1

        vertices = [None] * self.num_darts()
        edges = [None] * self.num_darts()
        faces = [None] * self.num_darts()
        for i in range(self.num_darts()):
            if self._active_darts[i]:
                vertices[perm[i]] = perm[self._vertices[i]]
                edges[perm[i]] = perm[self._edges[i]]
                faces[perm[i]] = perm[self._faces[i]]

        return RibbonGraph(vertices,edges,faces)

    #
    # Darts
    #

    def num_darts(self):
        r"""
        Returns the number of darts.
        """
        return self._num_darts

    def darts(self):
        r"""
        Return the list of darts
        """
        return [i for i in range(self._total_darts) if self._active_darts[i]]

    def num_vertices(self):
        r"""
        Returns the number of vertices.
        """
        return max(1,len(self._vertex_cycles))

    def vertex_perm(self):
        r"""
        Returns the permutation that define the vertices.
        """
        return self._vertices

    def vertex_orbit(self, i):
        r"""
        Return the orbit of ``i`` under the permutation that define the
        vertices.
        """
        if self._active_darts[i]:
            return perm_orbit(self._vertices,i)
        return None

    def vertices(self):
        r"""
        Return the list of vertices as cycles decomposition of the vertex
        permutation.
        """
        return self._vertex_cycles

    def dart_to_vertex(self,i):
        r"""
        Return the vertex on which the dart ``i`` is attached.
        """
        if self._active_darts[i]:
            return self._dart_to_vertex_index[i]
        raise ValueError("dart %d is not active" %i)

    #
    # Edges
    #

    def num_edges(self):
        r"""
        Returns the number of edges.
        """
        return len(self._edge_cycles)

    def edge_perm(self):
        r"""
        Return the permutation that define the edges.
        """
        return self._edges

    def edge_orbit(self, i):
        r"""
        Return the orbit of the dart ``i`` under the permutation that defines
        the edges.
        """
        if self._active_darts[i]:
            return perm_orbit(self._edges,i)
        return None

    def edges(self):
        r"""
        Return the set of edges.
        """
        return self._edge_cycles

    def dart_to_edge(self, i, orientation=False):
        r"""
        Returns the edge the darts ``i`` belongs to.

        If orientation is set to ``True`` then the output is a `2`-tuple
        ``(e,o)`` where ``e`` is the index of the edge and ``o`` is its
        orientation as ``+1`` or ``-1``.
        """
        if self._active_darts[i]:
            if not orientation:
                return self._dart_to_edge_index[i]
            j = self._dart_to_edge_index[i]
            if i == self._edge_cycles[j][0]:
                return (j,1)
            elif i == self._edge_cycles[j][1]:
                return (j,-1)
            else:
                raise ValueError("this should not happen!")
        raise ValueError("dart %d is not active" %i)

    #
    # Faces
    #

    def num_faces(self):
        r"""
        Return the number of faces.
        """
        return max(1,len(self._face_cycles))

    def face_perm(self):
        r"""
        Return the permutation that defines the face.
        """
        return self._faces

    def face_orbit(self, i):
        r"""
        Return the orbit of ``i`` under the permutation associated to faces.
        """
        if self._active_darts[i]:
            return perm_orbit(self._faces,i)
        return None

    def faces(self):
        r"""
        Return the list of faces.
        """
        return self._face_cycles

    def dart_to_face(self, i):
        if self._active_darts[i]:
            return self._dart_to_face_index[i]
        raise ValueError("dart {} is not active".format(i))

    def dual(self):
        r"""
        Returns the dual Ribbon graph.

        The *dual* ribbon graph of `(v,e,f)` is `(f^{-1}, e, v^{-1})`.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: r = RibbonGraph(edges='(0,1)',faces='(0)(1)'); r
            Ribbon graph with 1 vertex, 1 edge and 2 faces
            sage: r.dual()
            Ribbon graph with 2 vertices, 1 edge and 1 face
        """
        return RibbonGraph(
            vertices=perm_invert(self._faces),
            edges=self._edges,
            faces=perm_invert(self._vertices))

    #
    # euler characteristic
    #

    def euler_characteristic(self):
        r"""
        Returns the Euler characteristic of the embedded surface.

        The *Euler characteristic* of a surface complex is `V - E + F`, where
        `V` is the number of vertices, `E` the number of edges and `F` the
        number of faces.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: r = RibbonGraph(edges='(0,1)(2,3)(4,5)',faces='(0,2,4)(1)(3,5)')
            sage: r.euler_characteristic()
            2

            sage: r = RibbonGraph(edges='(0,1)(2,3)',faces='(0,2,1,3)')
            sage: r.euler_characteristic()
            0
        """
        return self.num_vertices() - self.num_edges() + self.num_faces()

    def is_plane(self):
        r"""
        Returns true if and only if the ribbon graph belongs in a sphere. In
        other words if it has genus 0.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: r = RibbonGraph(vertices='(0)(1)',edges='(0,1)')
            sage: r.is_plane()
            True

            sage: r = RibbonGraph(vertices='(0,1)',edges='(0,1)')
            sage: r.is_plane()
            True

            sage: r = RibbonGraph(edges='(0,1)(2,3)',faces='(0,2)(1,3)')
            sage: r.is_plane()
            True

            sage: r = RibbonGraph(edges='(0,1)(2,3)',faces='(0,2,1,3)')
            sage: r.is_plane()
            False
        """
        return self.euler_characteristic() == 2

    def is_plane_tree(self):
        r"""
        Returns True if and only if the ribbon graph is a planar tree. In other
        words, it has genus 0 and only one face.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: r = RibbonGraph(vertices='(0)(1)',edges='(0,1)')
            sage: r.is_plane_tree()
            True

            sage: r = RibbonGraph(vertices='(0)(1,2,4)(3)(5)',edges='(0,1)(2,3)(4,5)')
            sage: r.is_plane_tree()
            True

            sage: r = RibbonGraph(vertices='(0,1)',edges='(0,1)')
            sage: r.is_plane_tree()
            False
            sage: r.is_plane()
            True
        """
        return (self.num_faces() == 1 and self.genus() == 0)

    def is_triangulated(self):
        r"""
        Returns True if the surface is triangulated. In other words, faces
        consist only of the product of 3-cycles.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: r = RibbonGraph(edges='(0,1)(2,3)(4,5)',faces='(0,2,4)(1,5,3)')
            sage: r.is_triangulated()
            True

            sage: r = RibbonGraph(edges='(0,1)(2,3)',faces='(0,2,1,3)')
            sage: r.is_triangulated()
            False
        """
        return all(len(c) == 3 for c in self.faces())

    def genus(self):
        r"""
        Return the genus of the surface associated to this Ribbon graph.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: R = RibbonGraph(vertices='(1)(2)',edges='(1,2)')
            sage: R.genus()
            0

            sage: e='(1,3)(2,4)'
            sage: f='(1,2,3,4)'
            sage: RibbonGraph(edges=e,faces=f).genus()
            1

            sage: e='(1,3)(2,4)(5,7)(6,8)'
            sage: f='(1,2,3,4,5,6,7,8)'
            sage: RibbonGraph(edges=e,faces=f).genus()
            2

            sage: e='(1,3)(2,4)(5,7)(6,8)(9,11)(10,12)'
            sage: f='(1,2,3,4,5,6,7,8,9,10,11,12)'
            sage: RibbonGraph(edges=e,faces=f).genus()
            3
        """
        return 1 - self.euler_characteristic()//2

    #
    # cycles and fundamental group
    #

    def spanning_tree(self):
        r"""
        Return a spanning tree

        OUTPUT:

        - spanning tree as a DiGraph

        - remaining edges as 2-tuples ``(i,e[i])``

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: R = RibbonGraph('(1,2,3)','(1,2)(3,4)')
            sage: R
            Ribbon graph with 2 vertices, 2 edges and 2 faces
            sage: T,o = R.spanning_tree()
            sage: T
            Digraph on 2 vertices
            sage: T.edges()
            [(0, 1, (3, 4))]
            sage: o
            [(1, 2)]

            sage: R = RibbonGraph('(1,2,3)(4,5,6)','(1,2)(3,4)(5,6)')
            sage: R
            Ribbon graph with 2 vertices, 3 edges and 3 faces
            sage: T,o = R.spanning_tree()
            sage: T
            Digraph on 2 vertices
            sage: T.edges()
            [(0, 1, (3, 4))]
            sage: o
            [(1, 2), (5, 6)]

            sage: e = '(1,3)(5,7)(2,4)(6,8)'
            sage: f = '(1,2,3,4,5,6,7,8)'
            sage: R = RibbonGraph(edges=e, faces=f)
            sage: T,o = R.spanning_tree()
            sage: T
            Digraph on 1 vertex
            sage: o
            [[1, 3], [2, 4], [5, 7], [6, 8]]
        """
        from sage.graphs.digraph import DiGraph

        d = self.darts()
        v = self.vertices()
        e = self.edge_perm()

        if self.num_darts() == 0:
            return DiGraph(),[]
        if self.num_vertices() == 1:
            return DiGraph({0:[]}),self.edges()

        T = DiGraph()

        v0 = 0
        for root in v[0]:
            v1 = self.dart_to_vertex(e[root])
            if v1 != 0:
                break

        T.add_edge(v0,v1,(root,e[root]))
        o = []
        seen = set([v0,v1])       # seen vertices
        waiting = [e[root],root]  # waiting darts
        cc = []

        while waiting:
            ii = waiting.pop() # this is a dart
            v0 = self.dart_to_vertex(ii)
            seen.add(v0)
            for j in self.vertex_orbit(ii)[1:]:
                v1 = self.dart_to_vertex(e[j])
                if v1 in seen:
                    if j < e[j]:
                        o.append((j,e[j]))
                else:
                    T.add_edge(v0,v1,(j,e[j]))
                    waiting.append(e[j])
                    seen.add(v1)

        return T, sorted(o)

    def collapse(self, spanning_tree=None):
        r"""
        Return a ribbon graph callapsed along a spanning tree.

        The resulting graph is on the same surface as the preceding but has only
        one vertex. It could be used twice to provide a polygonal representation
        with one vertex and one face.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: R = RibbonGraph(vertices='(0,1,2,5)(3,7)(4,10,9)(6,11,12)(8,13)')
            sage: R.genus()
            1
            sage: R.num_vertices()
            5
            sage: R.num_edges()
            7
            sage: R.num_faces()
            2
            sage: R2 = R.collapse()
            sage: R2
            Ribbon graph with 1 vertex, 3 edges and 2 faces
            sage: R
            Ribbon graph with 5 vertices, 7 edges and 2 faces
            sage: R3 = R2.dual().collapse().dual()
            sage: R3
            Ribbon graph with 1 vertex, 2 edges and 1 face

        """
        from copy import deepcopy

        if spanning_tree is None:
            spanning_tree,_ = self.spanning_tree()

        darts_to_kill = set([])
        for v0,v1,e in spanning_tree.edges():
            darts_to_kill.add(e[0])
            darts_to_kill.add(e[1])

        new_edges = []
        for e in self.edges():
            if e[0] not in darts_to_kill:
                new_edges.append(e)

        new_faces = []
        for f in self.faces():
            ff = tuple(i for i in f if i not in darts_to_kill)
            if ff:
                new_faces.append(ff)

        return RibbonGraph(edges=tuple(new_edges), faces=tuple(new_faces))

    def boundaries(self):
        r"""
        Return the list of cycles which are boundaries.

        A cycle is a *boundary* if it bounds a face.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: r = RibbonGraph('(1,2,3)(4,5,6)','(1,2)(3,4)(5,6)')
            sage: r.boundaries()
            [[(1, 2)],  [(2, 1), (3, 4), (6, 5), (4, 3)], [(5, 6)]]

            sage: r = RibbonGraph('(1,2,3)(4,5)(6,7,8)',edges='(1,2)(3,4)(5,6)(7,8)')
            sage: r.boundaries()
            [[(1, 2)],  [(2, 1), (3, 4), (5, 6), (8, 7), (6, 5), (4, 3)], [(7, 8)]]
        """
        e = self.edge_perm()
        return sorted([[(i,e[i]) for i in f] for f in self.faces()])

    def cycle_basis(self, intersection=False, verbose=False):
        r"""
        Returns a base of oriented cycles of the Ribbon graph modulo boundaries.

        If ``intersection`` is set to True then the method also returns the
        intersection matrix of the cycles.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: r = RibbonGraph('(1,2,3)(4,5,6)','(1,2)(3,4)(5,6)')
            sage: r.cycle_basis()
            []

            sage: r = RibbonGraph('(1,2,3)(4,5)(6,7,8)',edges='(1,2)(3,4)(5,6)(7,8)')
            sage: r.cycle_basis()
            []

            sage: r = RibbonGraph('(1,4,5)(2,3)(6,7,8)',edges='(1,2)(3,4)(5,6)(7,8)')
            sage: r.cycle_basis()
            []

            sage: e = '(1,3)(2,4)(5,7)(6,8)'
            sage: f = '(1,2,3,4,5,6,7,8)'
            sage: r = RibbonGraph(edges=e,faces=f)
            sage: r.cycle_basis()
            [[[1, 3]], [[2, 4]], [[5, 7]], [[6, 8]]]

            sage: f = '(0,10,13)(6,17,11)(2,14,7)(15,12,3)(16,20,19)(18,1,9)(4,22,21)(23,8,5)'
            sage: e = tuple((i,i+1) for i in range(0,24,2))
            sage: r = RibbonGraph(edges=e,faces=f); r
            Ribbon graph with 2 vertices, 12 edges and 8 faces
            sage: c,m = r.cycle_basis(intersection=True)
            sage: c
            [[(0, 1), [4, 5]], [[8, 9]], [[12, 13]], [[14, 15], (1, 0)]]
            sage: m
            [ 0  1  0  0]
            [-1  0  0  0]
            [ 0  0  0  1]
            [ 0  0 -1  0]
        """
        T,o = self.spanning_tree()

        # build a Ribbon graph with one vertex and one face
        r = self.collapse(T).dual().collapse().dual()
        if T is None:
            return r.edges()

        if intersection:
            c = r.vertices()[0]
            M = len(c)
            I = []

        cycles = [] # the cycles
        for e in r.edges():
            if verbose:
                print("build cycle from edge %s between vertex v0=%d and v1=%d" %(str(e),self.dart_to_vertex(e[0]),self.dart_to_vertex(e[1])))

            # build the branch to the root from v0
            v0 = self.dart_to_vertex(e[0])
            if verbose:
                print(" build branch from v0=%d" % v0)
            p0 = []
            while v0 != 0:
                v0,_,e0 = T.incoming_edges(v0)[0] # (v_in,v_out,label)
                p0.append(e0)
                if verbose:
                    print(" add %d" % v0)
            if verbose:
                print(" branch is %s" % str(p0))
            # build the branch to the root from v1

            v1 = self.dart_to_vertex(e[1])
            if verbose:
                print(" build branch from v1=%d" % v1)
            p1 = []
            while v1 != 0:
                v1,_,e1 = T.incoming_edges(v1)[0]
                p1.append(e1)
                if verbose:
                    print(" add %d" % v1)
            if verbose:
                print(" branch is %s" % str(p1))
            # clean the branches by removing common part
            while p0 and p1 and p0[-1] == p1[-1]:
                if verbose:
                    print("find common element", p0[-1])
                p0.pop(-1)
                p1.pop(-1)

            # add the cycle to the list
            cycles.append((p0,e,p1))

            # compute algebraic intersection with preceding cycles
            if intersection:
                i = []
                for _,ee,_ in cycles:
                    if verbose:
                        print("compute intersection")
                    p_in = c.index(e[1])
                    p_out = (c.index(e[0]) - p_in) % M
                    q_in  = (c.index(ee[1]) - p_in) % M
                    q_out = (c.index(ee[0]) - p_in) % M
                    if verbose:
                        print("  after reduction: p_out = %d, q_in = %d, q_out = %d" % (p_out, q_in, q_out))

                    # compute intersection
                    # p_in = 0 and the others 3 are positive
                    if q_in < p_out and p_out < q_out:
                        i.append(1)
                    elif q_out < p_out and p_out < q_in:
                        i.append(-1)
                    else:
                        i.append(0)

                I.append(i)

        # make cycle as list
        cycles = [p0[::-1]+[e]+[c[::-1] for c in p1] for p0,e,p1 in cycles]

        if intersection:
            m = matrix(len(cycles))
            for j in range(len(I)):
                for jj in range(len(I[j])):
                    m[j,jj] = I[j][jj]
                    m[jj,j] = -I[j][jj]

            return cycles, m
        return cycles

    def is_cycle(self,c):
        r"""
        Test whether ``c`` is a cycle.

        A *path* is a sequence of oriented edges such that each edge starts
        where the preceding one ends. A *cycle* is a path which starts where it
        ends.
        """
        for i in range(len(c)-1):
            if self.dart_to_vertex(c[i][1]) != self.dart_to_vertex(c[i+1][0]):
                return False
        if self.dart_to_vertex(c[-1][1]) != self.dart_to_vertex(c[0][0]):
            return False
        return True

class RibbonGraphWithAngles(RibbonGraph):
    r"""
    A Ribbon graph with angles between edges

    Currently angles can only be *rational* multiples of pi.

    TODO:

    - allows any kind of angles by providing a sum for the total and considering
      each angle as a (projective) portion of the total angle.
    """
    def __init__(self, vertices=None, edges=None, faces=None, angles=None):
        r = RibbonGraph(vertices,edges,faces)
        RibbonGraph.__init__(self,r.vertex_perm(),r.edge_perm(),r.face_perm())

        if len(angles) != self.num_darts():
            raise ValueError("there are %d angles and %d darts" %(len(angles),self.num_darts()))
        self._angles = list(map(QQ,angles))
          # angle between a dart and its vertex-neighbour
          # (rational number as multiple of pi)

        self._total_angle = []
          # total angle around vertices
          # (integer which corresponds to a multiple of pi)

        for v in self.vertices():
            self._total_angle.append(sum(angles[i] for i in v))

        for f in self.faces():
            a = sum(self._angles[i] for i in f)
            if a != len(f)-2:
                raise ValueError("the angle of a face should be (nb_edges - 2) x pi")

    def angle_between_darts(self, d1, d2):
        r"""
        Return the angle between the darts ``d1`` and ``d2``
        """
        v = self.vertex_orbit(d1)
        if d2 not in v:
            raise ValueError("d1=%s and d2=%s are not at the same vertex" %(str(d1),str(d2)))

        a = 0
        i = 0
        while v[i] != d2:
            a += self._angles[v[i]]
            i += 1
        return a

    def angle_at_vertex(self,v):
        r"""
        Angle at a vertex (coefficient of pi)
        """
        return self._total_angle[v]

    def angle_at_vertices(self):
        r"""
        Return the list of angles at a vertex.
        """
        return self._total_angle

    def winding(self, c):
        r"""
        Return winding number along the cycle ``c``.

        This is NOT well defined because it depends on the way we choose to pass
        on the left or on the right at singularity.
        """
        a = 0
        for i in range(len(c)-1):
            d1 = c[i][1]
            d2 = c[i+1][0]
            if self.dart_to_vertex(d1) != self.dart_to_vertex(d2):
                raise ValueError("c is not a cycle")
            a += self.angle_between_darts(d1,d2)-1
        d1 = c[-1][1]
        d2 = c[0][0]
        if self.dart_to_vertex(d1) != self.dart_to_vertex(d2):
            raise ValueError("c is not a cycle")
        a += self.angle_between_darts(d1,d2)-1

        return a

    def holonomy_representation(self):
        r"""
        Return the holonomy representation in `SO(2)` as two lists.

        The first list correspond to cycles around vertices, while the second
        correspond to a cycle basis that generate homology.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: e = '(0,1)(2,3)'
            sage: f = '(0,2,1,3)'
            sage: a = [1/2,1/2,1/2,1/2]
            sage: r = RibbonGraphWithAngles(edges=e,faces=f,angles=a)
            sage: r.holonomy_representation()
            ([0], [0, 0])

        The standard cube::

            sage: e = tuple((i,i+1) for i in range(0,24,2))
            sage: f = '(0,20,7,10)(16,22,19,21)(2,9,5,23)(14,3,17,1)(12,8,15,11)(18,4,13,6)'
            sage: a = [1/2]*24
            sage: r = RibbonGraphWithAngles(edges=e,faces=f,angles=a)
            sage: r.holonomy_representation()
            ([3/2, 3/2, 3/2, 3/2, 3/2, 3/2, 3/2, 3/2], [])

        Two copies of a triangle::

            sage: e = '(0,1)(2,3)(4,5)'
            sage: f = '(0,2,4)(1,5,3)'
            sage: a = [1/2,1/6,1/3,1/3,1/6,1/2]
            sage: r = RibbonGraphWithAngles(edges=e,faces=f,angles=a)
            sage: r.holonomy_representation()
            ([1, 1/2, 1/2], [])

            sage: a = [1/3,7/15,1/5,1/5,7/15,1/3]
            sage: r = RibbonGraphWithAngles(edges=e,faces=f,angles=a)
            sage: r.holonomy_representation()
            ([2/3, 2/3, 2/3], [])
        """
        from sage.functions.other import floor

        l1 = []
        for c in range(self.num_vertices()):
            w = self.angle_at_vertex(c)
            l1.append(w - 2*floor(w/2))

        l2 = []
        for c in self.cycle_basis():
            w = self.winding(c)
            l2.append(w - 2*floor(w/2))

        return l1,l2

    def has_trivial_holonomy(self):
        r"""
        Test whether self has trivial holonomy representation

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: e = '(0,1)(2,3)'
            sage: f = '(0,2,1,3)'
            sage: a = [1/2,1/2,1/2,1/2]
            sage: r = RibbonGraphWithAngles(edges=e,faces=f,angles=a)
            sage: r.has_trivial_holonomy()
            True

            sage: e = '(0,1)(2,3)(4,5)'
            sage: f = '(0,2,4)(1,5,3)'
            sage: a = [1/3,7/15,1/5,1/5,7/15,1/3]
            sage: r = RibbonGraphWithAngles(edges=e,faces=f,angles=a)
            sage: r.has_trivial_holonomy()
            False
        """
        l1,l2 = self.holonomy_representation()
        return all(i==0 for i in l1) and all(i==0 for i in l2)

    def spin_parity(self,check=True,verbose=False):
        r"""
        Return the spin parity of the Ribbon graph with angles.

        The surface should be holonomy free and with odd multiple of 2 pi
        angles.

        EXAMPLES:

            sage: from surface_dynamics import *

        We first consider the case of the torus::

            sage: e = '(0,1)(2,3)'
            sage: f = '(0,2,1,3)'
            sage: a = [1/2,1/2,1/2,1/2]
            sage: r = RibbonGraphWithAngles(edges=e,faces=f,angles=a)
            sage: r.spin_parity()
            1

        Then the case of genus 2 surface (with an angle of 6pi)::

            sage: e = '(0,1)(2,3)(4,5)(6,7)'
            sage: f = '(0,2,4,3,6,1,7,5)'
            sage: a = [1/2,1/2,1,1/2,1/2,1,3/2,1/2]
            sage: r = RibbonGraphWithAngles(edges=e,faces=f,angles=a)
            sage: r.spin_parity()
            1

            sage: e = '(0,1)(2,3)(4,5)(6,7)'
            sage: f = '(0,2,4,6,1,3,5,7)'
            sage: a = [1/2,1/2,1,1,1,1,1/2,1/2]
            sage: r = RibbonGraphWithAngles(edges=e,faces=f,angles=a)
            sage: r.spin_parity()
            1

            sage: e = '(0,1)(2,3)(4,5)(6,7)'
            sage: f = '(0,2,4,6,1,3,5,7)'
            sage: a = [3/4]*8
            sage: r = RibbonGraphWithAngles(edges=e,faces=f,angles=a)
            sage: r.spin_parity()
            1

        In genus 3 two spin parities occur for one conical angle 10pi::

            sage: e = '(0,1)(2,3)(4,5)(6,7)(8,9)(10,11)'
            sage: f1 = '(0,4,6,8,10,2,1,9,11,5,7,3)'
            sage: f2 = '(0,4,6,8,10,2,1,5,7,9,11,3)'
            sage: a = [1/2,1/2,1/2,1/2] + [1]*8
            sage: r1 = RibbonGraphWithAngles(edges=e,faces=f1,angles=a)
            sage: r1.spin_parity()
            1
            sage: r2 = RibbonGraphWithAngles(edges=e,faces=f2,angles=a)
            sage: r2.spin_parity()
            0
        """
        from sage.rings.finite_rings.finite_field_constructor import GF
        # mod F2 we have: q(x+y) = B(x,y) + q(x) + q(y)

        if not self.has_trivial_holonomy():
            raise ValueError("the surface does not have trivial holonomy")
        if any((i+2)%4 for i in self.angle_at_vertices()):
            raise ValueError("each angle should be odd multiple of 2pi")

        GF2 = GF(2)

        c,M = self.cycle_basis(intersection=True)

        winding = []
        for cc in c:
            w = self.winding(cc)
            if w % 2 != 0:
                raise ValueError("fatal error ! each winding should be a multiple of 2")
            winding.append(GF2(w//2))

        if verbose:
            print("cycles with winding")
            for i in range(len(c)):
                print(c[i], winding[i])
            print("intersection matrix on Z")
            print(M)

        # compute a base change to get a symplectic basis
        _,P = M.symplectic_form()
        M = M.change_ring(GF2)
        P = P.change_ring(GF2)
        if verbose:
            print("base change for symplectic basis on GF(2)")
            print(P)

        g = self.genus()

        s = GF2(0)
        for i in range(g):
            # 1. computation of q(P.row(i))
            a = P.row(i)
            a_indices = [j for j in range(2*g) if a[j] != 0]
            ## winding + nb_components
            t_a = sum(winding[i]+1 for i in a_indices)
            ## self intersection
            for j1 in range(len(a_indices)):
                for j2 in range(j1+1,len(a_indices)):
                    t_a += M[a_indices[j1],a_indices[j2]]

            # 2. computation of q(P.row(g+i))
            b = P.row(g+i)
            b_indices = [j for j in range(2*g) if b[j] != 0]
            ## winding + nb_components
            t_b = sum(winding[i]+1 for i in b_indices)
            ## self intersection
            for j1 in range(len(b_indices)):
                for j2 in range(j1+1,len(b_indices)):
                    t_b += M[b_indices[j1],b_indices[j2]]

            # 3. add to s the contribution of the couple
            if verbose:
                print("contribution from %d is %d * %d = %d" % (i, t_a, t_b, t_a * t_b))
            s += t_a*t_b

        return s

def angle(v):
    r"""
    Return the argument of the vector ``v``.
    """
    from math import acos,asin,sqrt,pi
    x = float(v[0])
    y = float(v[1])
    r = sqrt(x**2 + y**2)
    if abs(x) >= abs(y):
        if x >= 0:
            return asin(y / r) / pi
        else:
            return -asin(y / r) / pi
    else:
        if y >= 0:
            return acos(x / r) / pi
        else:
            return -acos(x / r) / pi

class RibbonGraphWithHolonomies(RibbonGraph):
    r"""
    A Ribbon graph with holonomies.

    For now
    """
    def __init__(self, vertices=None, edges=None, faces=None, holonomies=None):
        r = RibbonGraph(vertices,edges,faces)
        RibbonGraph.__init__(self,r.vertex_perm(),r.edge_perm(),r.face_perm())

        if len(holonomies) != self.num_darts():
            raise ValueError("there are %d angles and %d darts" %(len(angles),self.num_darts()))

        from sage.modules.free_module import FreeModule
        V = FreeModule(ZZ,2)
        self._holonomies = list(map(V, holonomies))

        #self._angles = map(angle, self._holonomies)







r"""
Reduced permutations

A reduced (generalized) permutation is better suited to study strata of Abelian
(or quadratic) holomorphic forms on Riemann surfaces. The Rauzy diagram is an
invariant of such a component. Corentin Boissy proved the identification of
Rauzy diagrams with connected components of stratas. But the geometry of the
diagram and the relation with the strata is not yet totally understood.

AUTHORS:

- Vincent Delecroix (2000-09-29): initial version

TESTS::

    sage: from surface_dynamics.interval_exchanges.reduced import ReducedPermutationIET
    sage: ReducedPermutationIET([['a','b'],['b','a']])
    a b
    b a
    sage: ReducedPermutationIET([[1,2,3],[3,1,2]])
    1 2 3
    3 1 2
    sage: from surface_dynamics.interval_exchanges.reduced import ReducedPermutationLI
    sage: ReducedPermutationLI([[1,1],[2,2,3,3,4,4]])
    1 1
    2 2 3 3 4 4
    sage: ReducedPermutationLI([['a','a','b','b','c','c'],['d','d']])
    a a b b c c
    d d
    sage: from surface_dynamics.interval_exchanges.reduced import FlippedReducedPermutationIET
    sage: FlippedReducedPermutationIET([[1,2,3],[3,2,1]],flips=[1,2])
    -1 -2  3
     3 -2 -1
    sage: FlippedReducedPermutationIET([['a','b','c'],['b','c','a']],flips='b')
     a -b  c
    -b  c  a
    sage: from surface_dynamics.interval_exchanges.reduced import FlippedReducedPermutationLI
    sage: FlippedReducedPermutationLI([[1,1],[2,2,3,3,4,4]], flips=[1,4])
    -1 -1
     2  2  3  3 -4 -4
    sage: FlippedReducedPermutationLI([['a','a','b','b'],['c','c']],flips='ac')
    -a -a  b  b
    -c -c
    sage: from surface_dynamics.interval_exchanges.reduced import ReducedRauzyDiagram
    sage: p = ReducedPermutationIET([[1,2,3],[3,2,1]])
    sage: d = ReducedRauzyDiagram(p)
"""
#*****************************************************************************
#       Copyright (C) 2008 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

from __future__ import print_function, absolute_import, division
from six.moves import range, map, filter, zip

from sage.structure.sage_object import SageObject

from copy import copy

from sage.combinat.words.alphabet import Alphabet
from sage.rings.integer import Integer

from .template import OrientablePermutationIET, OrientablePermutationLI   # permutations
from .template import FlippedPermutationIET, FlippedPermutationLI         # flipped permutations
from .template import RauzyDiagram, FlippedRauzyDiagram

from .template import interval_conversion, side_conversion

class ReducedPermutation(SageObject) :
    r"""
    Template for reduced objects.

    .. warning::

        Internal class! Do not use directly!
   """
    def __getitem__(self, i):
        r"""
        TESTS::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation('a b', 'b a', reduced=True)
            sage: p[0]
            ['a', 'b']
            sage: p[1]
            ['b', 'a']
            sage: p.alphabet([0,1])
            sage: p[0]
            [0, 1]
            sage: p[1]
            [1, 0]
        """
        return self.list()[i]

def ReducedPermutationsIET_iterator(
    nintervals=None,
    irreducible=True,
    alphabet=None):
    r"""
    Returns an iterator over reduced permutations

    INPUT:

    - ``nintervals`` - integer or None

    - ``irreducible`` - boolean

    - ``alphabet`` - something that should be converted to an alphabet
      of at least nintervals letters

    TESTS::

        sage: from surface_dynamics import *

        sage: for p in iet.Permutations_iterator(3,reduced=True,alphabet="abc"):
        ....:     print(p)  #indirect doctest
        a b c
        b c a
        a b c
        c a b
        a b c
        c b a
    """
    from sage.combinat.permutation import Permutations

    if irreducible is False:
        if nintervals is None:
            raise NotImplementedError("choose a number of intervals")
        else:
            assert(isinstance(nintervals,(int,Integer)))
            assert(nintervals > 0)

            a0 = range(1,nintervals+1)
            f = lambda x: ReducedPermutationIET([a0,list(x)],
                alphabet=alphabet, reduced=True)
            return map(f, Permutations(nintervals))
    else:
        return filter(lambda x: x.is_irreducible(),
        ReducedPermutationsIET_iterator(nintervals,False,alphabet))

class ReducedPermutationIET(ReducedPermutation, OrientablePermutationIET):
    """
    Reduced permutation from iet

    Permutation from iet without numerotation of intervals. For initialization,
    you should use GeneralizedPermutation which is the class factory for all
    permutation types.

    EXAMPLES::

        sage: from surface_dynamics import *

    Equality testing (no equality of letters but just of ordering)::

        sage: p = iet.Permutation('a b c', 'c b a', reduced = True)
        sage: q = iet.Permutation('p q r', 'r q p', reduced = True)
        sage: p == q
        True

    Reducibility testing::

        sage: p = iet.Permutation('a b c', 'c b a', reduced = True)
        sage: p.is_irreducible()
        True

    ::

        sage: q = iet.Permutation('a b c d', 'b a d c', reduced = True)
        sage: q.is_irreducible()
        False


    Rauzy movability and Rauzy move::

        sage: p = iet.Permutation('a b c', 'c b a', reduced = True)
        sage: p.has_rauzy_move(1)
        True
        sage: p.rauzy_move(1)
        a b c
        b c a

    Rauzy diagrams::

        sage: p = iet.Permutation('a b c d', 'd a b c')
        sage: p_red = iet.Permutation('a b c d', 'd a b c', reduced = True)
        sage: d = p.rauzy_diagram()
        sage: d_red = p_red.rauzy_diagram()
        sage: p.rauzy_move(0) in d
        True
        sage: d.cardinality()
        12
        sage: d_red.cardinality()
        6
    """
    def list(self):
        r"""
        Returns a list of two list that represents the permutation.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.GeneralizedPermutation('a b','b a',reduced=True)
            sage: p.list() == [['a', 'b'], ['b', 'a']]
            True
            sage: p = iet.GeneralizedPermutation('a a','b b',reduced=True)
            sage: p.list() == [['a', 'a'], ['b', 'b']]
            True
        """
        return [
            list(map(lambda x: self._alphabet.unrank(x), range(len(self._twin[0])))),
            list(map(lambda x: self._alphabet.unrank(x), self._twin[1]))]

    def rauzy_move_relabel(self, winner, side='right'):
        r"""
        Returns the relabelization obtained from this move.

        EXAMPLE::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation('a b c d','d c b a')
            sage: q = p.reduced()
            sage: p_t = p.rauzy_move('t')
            sage: q_t = q.rauzy_move('t')
            sage: s_t = q.rauzy_move_relabel('t')
            sage: print(s_t)
            a->a, b->b, c->c, d->d
            sage: list(map(s_t, p_t[0])) == list(map(Word, q_t[0]))
            True
            sage: list(map(s_t, p_t[1])) == list(map(Word, q_t[1]))
            True
            sage: p_b = p.rauzy_move('b')
            sage: q_b = q.rauzy_move('b')
            sage: s_b = q.rauzy_move_relabel('b')
            sage: print(s_b)
            a->a, b->d, c->b, d->c
            sage: list(map(s_b, q_b[0])) == list(map(Word, p_b[0]))
            True
            sage: list(map(s_b, q_b[1])) == list(map(Word, p_b[1]))
            True
        """
        from surface_dynamics.interval_exchanges.labelled import LabelledPermutationIET
        from sage.combinat.words.morphism import WordMorphism

        winner = interval_conversion(winner)
        side = side_conversion(side)

        p = LabelledPermutationIET(self.list())

        l0_q = p.rauzy_move(winner, side).list()[0]

        d = dict([(self._alphabet[i],l0_q[i]) for i in range(len(self))])

        return WordMorphism(d)

    def rauzy_diagram(self, extended=False, **kwds):
        r"""
        Returns a Rauzy diagram associated to this permutation

        INPUT:

        - ``extended`` - boolean (default: False) - if True return extended
          Rauzy diagram

        OUTPUT:

        A Rauzy diagram

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation('a b c d', 'd a b c',reduced=True)
            sage: d = p.rauzy_diagram()
            sage: p.rauzy_move(0) in d
            True
            sage: p.rauzy_move(1) in d
            True

        For more information, try help RauzyDiagram
        """
        options = kwds
        if extended:
            options['symmetric'] = True
        return ReducedRauzyDiagram(self,**options)

    def rauzy_class_cardinality(self, extended=False):
        r"""
        Cardinality of Rauzy diagram

        As proved in [Del10], there exists a closed formula for the cardinality
        of Rauzy diagrams. This function uses the formula without any
        explicit computation of the rauzy class.

        INPUT:

        - ``extended`` - boolean (default: False) - return cardinality for
          extended Rauzy diagrams

        EXAMPLES::

            sage: from surface_dynamics import *

        Examples for permutations such that the suspensions are tori::

            sage: p=iet.Permutation('a b','b a',reduced=True)
            sage: p.stratum()
            H_1(0)
            sage: p.rauzy_diagram()
            Rauzy diagram with 1 permutation
            sage: p.rauzy_class_cardinality()
            1

            sage: p = iet.Permutation('a 1 b','b 1 a',reduced=True)
            sage: p.stratum()
            H_1(0^2)
            sage: p.rauzy_diagram()
            Rauzy diagram with 3 permutations
            sage: p.rauzy_class_cardinality()
            3

            sage: p = iet.Permutation('a 1 2 b','b 1 2 a',reduced=True)
            sage: p.stratum()
            H_1(0^3)
            sage: p.rauzy_diagram()
            Rauzy diagram with 6 permutations
            sage: p.rauzy_class_cardinality()
            6

            sage: p = iet.Permutation('a 1 2 3 b','b 1 2 3 a',reduced=True)
            sage: p.rauzy_class_cardinality()
            10
            sage: p = iet.Permutation('a 1 2 3 4 b','b 1 2 3 4 a',reduced=True)
            sage: p.rauzy_class_cardinality()
            15

        You should have recognize the sequence 1, 3, 6, 10, 15... which is the
        sequence with general term the binomial ``n(n+1)/2``.

        An example of extended Rauzy diagram which is different from Rauzy
        diagram::

            sage: p = iet.Permutation('a b c d e f g','g c b f e d a',reduced=True)
            sage: p.marked_profile()
            2o4 [4, 2]
            sage: pp = p.left_right_inverse()
            sage: pp.marked_profile()
            4o2 [4, 2]
            sage: p.rauzy_class_cardinality()
            261
            sage: pp.rauzy_class_cardinality()
            509
            sage: p.rauzy_class_cardinality(extended=True)
            770
            sage: 261 + 509 == 770
            True

        And one can check that this the algorithm for cardinality is True::

            sage: p.rauzy_diagram()
            Rauzy diagram with 261 permutations
            sage: pp.rauzy_diagram()
            Rauzy diagram with 509 permutations
            sage: p.rauzy_diagram(extended=True)
            Rauzy diagram with 770 permutations

        And we end by an example of Rauzy diagram associated to an hyperelliptic
        component::

            sage: p = iet.Permutation('a b c d e 0 f','f e d c b 0 a',reduced=True)
            sage: p.rauzy_diagram()
            Rauzy diagram with 37 permutations
            sage: p.rauzy_class_cardinality()
            37
            sage: p.rauzy_diagram(extended=True)
            Rauzy diagram with 254 permutations
            sage: p.rauzy_class_cardinality(extended=True)
            254
        """
        from .rauzy_class_cardinality import gamma_irr,delta_irr
        from sage.arith.all import binomial

        s = self.stratum()
        mp = self.marked_profile()
        p = mp.partition()

        if extended:
            return self.stratum_component().rauzy_class_cardinality()

        if s.is_connected():  # easy stuff
            return gamma_irr(p,mp.left())

        cc = self.stratum_component()
        zeros = s.zeros(fake_zeros=False)
        g = s.genus()
        if zeros == [2*g-2] or zeros == [g-1,g-1]:  # hyp component + others
            if cc.spin() == s.hyperelliptic_component().spin():
                d = len(self) # number of intervals of self
                if g == 1: # genus 1 is particular
                    nb_hyp = binomial(d,2)
                else:
                    k = s.nb_fake_zeros()
                    d -= k
                    if extended:
                        nb_hyp = binomial(d+k,k) * (2*d-1) + binomial(d+k-1,k-1) * d
                    elif mp.left() == 1:
                        nb_hyp = binomial(d+k,k-1) * (2**(d-1)-1+d)
                    elif mp.left() == d-1:
                        nb_hyp = binomial(d+k,k) * (2**(d-1)-1)
                    else:
                        raise ValueError("this should not happen")

                if cc == s.hyperelliptic_component(): # hyp
                    return nb_hyp

                if s.number_of_components() == 2: # other is alone
                    N = gamma_irr(p,mp.left())

                else: # others are odd and even
                    t = (-1)**(1-cc.spin())
                    N = (gamma_irr(p,mp.left()) + t * delta_irr(p,mp.left())) // 2

                return N - nb_hyp

            else: # cc.spin() != s.hyperelliptic_component().spin()
                t = (-1)**(1-cc.spin())
                return (gamma_irr(p,mp.left()) + t * delta_irr(p,mp.left())) // 2

        else:   # odd and even components
            t = (-1)**(1-cc.spin())  # 1 if odd and -1 if even
            return (gamma_irr(p,mp.left()) + t * delta_irr(p,mp.left())) // 2

class ReducedPermutationLI(ReducedPermutation, OrientablePermutationLI):
    r"""
    Reduced quadratic (or generalized) permutation.

    EXAMPLES::

        sage: from surface_dynamics import *

    Reducibility testing::

        sage: p = iet.GeneralizedPermutation('a b b', 'c c a', reduced = True)
        sage: p.is_irreducible()
        True

    ::

        sage: p = iet.GeneralizedPermutation('a b c a', 'b d d c', reduced = True)
        sage: p.is_irreducible()
        False
        sage: test, decomposition = p.is_irreducible(return_decomposition = True)
        sage: test
        False
        sage: decomposition
        (['a'], ['c', 'a'], [], ['c'])

    Rauzy movavability and Rauzy move::

        sage: p = iet.GeneralizedPermutation('a b b', 'c c a', reduced = True)
        sage: p.has_rauzy_move(0)
        True
        sage: p.rauzy_move(0)
        a a b b
        c c
        sage: p.rauzy_move(0).has_rauzy_move(0)
        False
        sage: p.rauzy_move(1)
        a b b
        c c a

    Rauzy diagrams::

        sage: p_red = iet.GeneralizedPermutation('a b b', 'c c a', reduced = True)
        sage: d_red = p_red.rauzy_diagram()
        sage: d_red.cardinality()
        4
    """
    def list(self) :
        r"""
        The permutations as a list of two lists.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.GeneralizedPermutation('a b b', 'c c a', reduced = True)
            sage: list(p)
            [['a', 'b', 'b'], ['c', 'c', 'a']]
        """
        i_a = 0
        l = [[False]*len(self._twin[0]),[False]*len(self._twin[1])]
        # False means empty here
        for i in range(2) :
            for j in range(len(l[i])) :
                if  l[i][j] is False :
                    l[i][j] = self._alphabet[i_a]
                    l[self._twin[i][j][0]][self._twin[i][j][1]] = self._alphabet[i_a]
                    i_a += 1
        return l

    def rauzy_diagram(self, **kargs):
        r"""
        Returns the associated Rauzy diagram.

        The Rauzy diagram of a permutation corresponds to all permutations
        that we could obtain from this one by Rauzy move. The set obtained
        is a labelled Graph. The label of vertices being 0 or 1 depending
        on the type.

        OUTPUT:

        Rauzy diagram -- the graph of permutations obtained by rauzy induction

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation('a b c d', 'd a b c')
            sage: d = p.rauzy_diagram()
        """
        return ReducedRauzyDiagram(self, **kargs)

def labelize_flip(couple):
    r"""
    Returns a string from a 2-uple couple of the form (name, flip).

    TESTS::

        sage: from surface_dynamics.interval_exchanges.reduced import labelize_flip
        sage: labelize_flip((4,1))
        ' 4'
        sage: labelize_flip(('a',-1))
        '-a'
    """
    if couple[1] == -1: return '-' + str(couple[0])
    return ' ' + str(couple[0])

FlippedReducedPermutation = ReducedPermutation

class FlippedReducedPermutationIET(
    FlippedReducedPermutation,
    FlippedPermutationIET):
    r"""
    Flipped Reduced Permutation from iet

    EXAMPLES::

        sage: from surface_dynamics import *

        sage: p = iet.Permutation('a b c', 'c b a', flips=['a'], reduced=True)
        sage: p.rauzy_move(1)
        -a -b  c
        -a  c -b

    TESTS::

        sage: p = iet.Permutation('a b','b a',flips=['a'])
        sage: p == loads(dumps(p))
        True

        sage: p = iet.Permutation('a b c', 'c a b', flips=['b'], reduced=True)
        sage: for q in sorted(p.rauzy_diagram()):
        ....:     print('%s\n********' % q)
        -a  b -c
         b -a -c
        ********
         a -b -c
        -b  a -c
        ********
         a -b  c
         c  a -b
        ********
         a  b -c
         b -c  a
        ********
         a -b  c
         c -b  a
        ********
    """
    def list(self, flips=False):
        r"""
        Returns a list representation of self.

        INPUT:

        - ``flips`` - boolean (default: False) if True the output contains
           2-uple of (label, flip)

        EXAMPLES::

            sage: from surface_dynamics import iet

            sage: p = iet.Permutation('a b','b a',reduced=True,flips='b')
            sage: p.list(flips=True)
            [[('a', 1), ('b', -1)], [('b', -1), ('a', 1)]]
            sage: p.list(flips=False)
            [['a', 'b'], ['b', 'a']]
            sage: p.alphabet([0,1])
            sage: p.list(flips=True)
            [[(0, 1), (1, -1)], [(1, -1), (0, 1)]]
            sage: p.list(flips=False)
            [[0, 1], [1, 0]]

        One can recover the initial permutation from this list::

            sage: p = iet.Permutation('a b','b a',reduced=True,flips='a')
            sage: iet.Permutation(p.list(), flips=p.flips(), reduced=True) == p
            True
        """
        if flips:
            a0 = list(zip(map(self.alphabet().unrank, range(0,len(self))), self._flips[0]))
            a1 = list(zip(map(self.alphabet().unrank, self._twin[1]), self._flips[1]))

        else:
            a0 = list(map(self.alphabet().unrank, range(0,len(self))))
            a1 = list(map(self.alphabet().unrank, self._twin[1]))

        return [a0,a1]

    def rauzy_diagram(self, **kargs):
        r"""
        Returns the associated Rauzy diagram.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation('a b','b a',reduced=True,flips='a')
            sage: r = p.rauzy_diagram()
            sage: p in r
            True
        """
        return FlippedReducedRauzyDiagram(self, **kargs)

class FlippedReducedPermutationLI(
    FlippedReducedPermutation,
    FlippedPermutationLI):
    r"""
    Flipped Reduced Permutation from li

    EXAMPLES:

            sage: from surface_dynamics import *

    Creation using the GeneralizedPermutation function::

        sage: p = iet.GeneralizedPermutation('a a b', 'b c c', reduced=True, flips='a')
    """
    def list(self, flips=False):
        r"""
        Returns a list representation of self.

        INPUT:

        - ``flips`` - boolean (default: False) return the list with flips

        EXAMPLES:

            sage: from surface_dynamics import *

        ::

            sage: p = iet.GeneralizedPermutation('a a','b b',reduced=True,flips='a')
            sage: p.list(flips=True)
            [[('a', -1), ('a', -1)], [('b', 1), ('b', 1)]]
            sage: p.list(flips=False)
            [['a', 'a'], ['b', 'b']]

            sage: p = iet.GeneralizedPermutation('a a b','b c c',reduced=True,flips='abc')
            sage: p.list(flips=True)
            [[('a', -1), ('a', -1), ('b', -1)], [('b', -1), ('c', -1), ('c', -1)]]
            sage: p.list(flips=False)
            [['a', 'a', 'b'], ['b', 'c', 'c']]

        One can rebuild the permutation from the list::

            sage: p = iet.GeneralizedPermutation('a a b','b c c',flips='a',reduced=True)
            sage: iet.GeneralizedPermutation(p.list(),flips=p.flips(),reduced=True) == p
            True
        """
        i_a = 0
        l = [[False]*len(self._twin[0]),[False]*len(self._twin[1])]

        if flips:
            for i in range(2):  # False means empty here
                for j in range(len(l[i])):
                   if  l[i][j] is False:
                        l[i][j] = (self._alphabet.unrank(i_a), self._flips[i][j])
                        l[self._twin[i][j][0]][self._twin[i][j][1]] = l[i][j]
                        i_a += 1

        else:
            for i in range(2):  # False means empty here
                for j in range(len(l[i])):
                   if  l[i][j] is False:
                        l[i][j] = self._alphabet.unrank(i_a)
                        l[self._twin[i][j][0]][self._twin[i][j][1]] = l[i][j]
                        i_a += 1
        return l

    def rauzy_diagram(self, **kargs):
        r"""
        Returns the associated Rauzy diagram.

        For more explanation and a list of arguments try help(iet.RauzyDiagram)

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.GeneralizedPermutation('a a b','c c b',reduced=True)
            sage: r = p.rauzy_diagram()
            sage: p in r
            True
        """
        return FlippedReducedRauzyDiagram(self, **kargs)

class ReducedRauzyDiagram(RauzyDiagram):
    r"""
    Rauzy diagram of reduced permutations
    """
    def _permutation_to_vertex(self, p):
        r"""
        The necessary data to store the permutation.

        TESTS::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation('a b c','c b a',reduced=True)   #indirect doctest
            sage: r = p.rauzy_diagram()
            sage: p in r
            True
        """
        return (tuple(p._twin[0]), tuple(p._twin[1]))

    def _set_element(self, data=None):
        r"""
        Sets self._element with data.

        TESTS::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation('a b c','c b a',reduced=True)
            sage: r = p.rauzy_diagram()
            sage: p in r   #indirect doctest
            True
        """
        self._element._twin = [list(data[0]), list(data[1])]

class FlippedReducedRauzyDiagram(FlippedRauzyDiagram, ReducedRauzyDiagram):
    r"""
    Rauzy diagram of flipped reduced permutations.

    TESTS::

        sage: from surface_dynamics import *

        sage: p = iet.Permutation('a b c', 'c a b', flips=['b'], reduced=True)
        sage: r = p.rauzy_diagram()
        sage: r
        Rauzy diagram with 5 permutations
        sage: p in r
        True
        sage: p.rauzy_move('t','r') in r
        True
        sage: p.rauzy_move('b','r') in r
        True

        sage: p = iet.GeneralizedPermutation('a b b','c c a',flips='a',reduced=True)
        sage: r = p.rauzy_diagram()
        Traceback (most recent call last):
        ...
        NotImplementedError: irreducibility test not implemented for generalized permutations with flips
    """
    def _permutation_to_vertex(self, p):
        return (tuple(p._twin[0]), tuple(p._twin[1]),
                tuple(p._flips[0]), tuple(p._flips[1]))

    def _set_element(self, data=None):
        self._element._twin = [list(data[0]), list(data[1])]
        self._element._flips = [list(data[2]), list(data[3])]



#coding: utf-8
r"""
Some utility function for representations of finite groups

Most of the functions are just GAP wrappers.
"""
# *************************************************************************
# Copyright (C) 2015-2016 Charles Fougeron <charlesfougeron@gmail.com>
#                         Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# *************************************************************************

from sage.libs.gap.libgap import libgap

def real_characters(G):
    r"""
    Return a pair ``(table of characters, character degrees)`` for the
    group ``G``.

    OUPUT:

    - table of characters - a list of characters. Each character is represented
      as the list of its values on conjugacy classes. The order of conjugacy
      classes is the same as the one returned by GAP.

    - degrees - the list of degrees of the characters

    EXAMPLES::

        sage: from surface_dynamics.misc.group_representation import real_characters
        sage: T, deg = real_characters(AlternatingGroup(5))
        sage: T
        [(1, 1, 1, 1, 1),
         (3, -1, 0, -E(5) - E(5)^4, -E(5)^2 - E(5)^3),
         (3, -1, 0, -E(5)^2 - E(5)^3, -E(5) - E(5)^4),
         (4, 0, 1, -1, -1),
         (5, 1, -1, 0, 0)]
        sage: set(parent(x) for chi in T for x in chi)
        {Universal Cyclotomic Field}
        sage: deg
        [1, 3, 3, 4, 5]

        sage: T, deg = real_characters(CyclicPermutationGroup(6))
        sage: T
        [(1, 1, 1, 1, 1, 1),
         (1, -1, 1, -1, 1, -1),
         (2, -1, -1, 2, -1, -1),
         (2, 1, -1, -2, -1, 1)]
        sage: deg
        [1, 1, 1, 1]
    """
    from sage.rings.universal_cyclotomic_field import UniversalCyclotomicField

    G = libgap(G)
    UCF = UniversalCyclotomicField()
    n = G.ConjugacyClasses().Length()
    Tgap = G.CharacterTable().Irr()
    degrees = [chi.Degree().sage() for chi in Tgap]
    Tgap = [tuple(UCF(chi[j]) for j in range(n)) for chi in Tgap]

    real_T = []
    real_degrees = []
    seen = set()
    for i,chi in enumerate(Tgap):
        if chi in seen:
            continue

        real_degrees.append(degrees[i])
        if all(x.is_real() for x in chi):
            real_T.append(chi)
            seen.add(chi)
        else:
            seen.add(chi)
            chi_bar = tuple(z.conjugate() for z in chi)
            seen.add(chi_bar)
            real_T.append(tuple(chi[j] + chi[j].conjugate() for j in range(n)))

    return (real_T, real_degrees)

def conjugacy_class_matrix(cl, d):
    r"""
    Return the matrix associated to a given conjugacy class of a permutation
    group.

    The result is a `d \times d` matrix that is invariant under conjugation by
    the group action on `\ZZ^d`. It is used to produce the projection matrices
    on the isotypic subspaces.

    EXAMPLES::
        
        sage: from surface_dynamics.misc.group_representation import conjugacy_class_matrix

        sage: G = QuaternionGroup()
        sage: Ggap = libgap(G)
        sage: cls = Ggap.ConjugacyClasses()
        sage: m = conjugacy_class_matrix(cls[2], 8)
        sage: m
        [0 0 1 0 0 0 0 0]
        [0 0 0 1 0 0 0 0]
        [1 0 0 0 0 0 0 0]
        [0 1 0 0 0 0 0 0]
        [0 0 0 0 0 0 1 0]
        [0 0 0 0 0 0 0 1]
        [0 0 0 0 1 0 0 0]
        [0 0 0 0 0 1 0 0]

        sage: for cl in cls:
        ....:      m = conjugacy_class_matrix(cl, 8)
        ....:      for g in G:
        ....:          g = g.matrix()
        ....:          assert g*m*~g == m

        sage: A = AlternatingGroup(5)
        sage: Agap = libgap(A)
        sage: Ggap = Agap.Action(Agap)
        sage: gens = libgap.GeneratorsOfGroup(Ggap).sage()
        sage: G = PermutationGroup(gens)
        sage: cls = Ggap.ConjugacyClasses()
        sage: for cl in cls:
        ....:      m = conjugacy_class_matrix(cls[2], 60)
        ....:      for _ in range(20):
        ....:           g = G.random_element().matrix()
        ....:           assert g*m*~g == m
    """
    res = [[0]*d for _ in range(d)]

    for p in cl.AsList():
        p = [i-1 for i in libgap.ListPerm(p, d).sage()]
        for k in range(d):
            res[k][p[k]] += 1

    from sage.matrix.constructor import matrix
    return matrix(res)

def isotypic_projection_matrix(G, d, chi, deg, conj_mats=None):
    r"""
    Return an isotypic projection matrix

    INPUT:

    - ``G`` -- a permutation group

    - ``d`` -- (integer) the domain of the group is `\{1, 2, \ldots, d\}`

    - ``chi`` -- (tuple) real or complex character

    - ``deg`` -- (integer) degree of the character

    Recall the formula for the projection as given in Theorem 8 in [Ser]_. If
    `G` is a permutation group, then
    
    .. MATH::
    
        \pi_\chi = \sum_{g \in G} \overline_{\chi(g)} g

    REFERENCES::

    .. [Ser] J.-P. Serre, "Représentation des groupes finis."

    EXAMPLES::

        sage: from surface_dynamics.misc.group_representation import real_characters, isotypic_projection_matrix
        sage: G = AlternatingGroup(5)
        sage: T,deg = real_characters(G)
        sage: isotypic_projection_matrix(G, 5, T[0], deg[0])
        [1/5 1/5 1/5 1/5 1/5]
        [1/5 1/5 1/5 1/5 1/5]
        [1/5 1/5 1/5 1/5 1/5]
        [1/5 1/5 1/5 1/5 1/5]
        [1/5 1/5 1/5 1/5 1/5]
        sage: isotypic_projection_matrix(G, 5, T[1], deg[1])
        [0 0 0 0 0]
        [0 0 0 0 0]
        [0 0 0 0 0]
        [0 0 0 0 0]
        [0 0 0 0 0]
        sage: isotypic_projection_matrix(G, 5, T[2], deg[2])
        [0 0 0 0 0]
        [0 0 0 0 0]
        [0 0 0 0 0]
        [0 0 0 0 0]
        [0 0 0 0 0]
        sage: isotypic_projection_matrix(G, 5, T[3], deg[3])
        [ 4/5 -1/5 -1/5 -1/5 -1/5]
        [-1/5  4/5 -1/5 -1/5 -1/5]
        [-1/5 -1/5  4/5 -1/5 -1/5]
        [-1/5 -1/5 -1/5  4/5 -1/5]
        [-1/5 -1/5 -1/5 -1/5  4/5]
        sage: isotypic_projection_matrix(G, 5, T[4], deg[4])
        [0 0 0 0 0]
        [0 0 0 0 0]
        [0 0 0 0 0]
        [0 0 0 0 0]
        [0 0 0 0 0]

        sage: sum(isotypic_projection_matrix(G, 5, T[i], deg[i]) for i in range(5)).is_one()
        True
    """
    from sage.matrix.special import zero_matrix
    res = zero_matrix(d)

    Ggap = libgap(G)

    for t,cl in enumerate(Ggap.ConjugacyClasses()):
        if conj_mats is None:
            m = conjugacy_class_matrix(cl, d)
        else:
            m = conj_mats[t]
        res += chi[t] * conjugacy_class_matrix(cl,d)

    return deg / G.cardinality() * res

def real_isotypic_projection_matrices(G, d):
    r"""
    Return the real projections

    EXAMPLES::

        sage: from surface_dynamics.misc.group_representation import real_isotypic_projection_matrices

        sage: G = AlternatingGroup(5)
        sage: mats = real_isotypic_projection_matrices(G, 6)
        sage: sum(mats).is_one()
        True

        sage: G = CyclicPermutationGroup(6)
        sage: mats = real_isotypic_projection_matrices(G, 6)
        sage: sum(mats).is_one()
        True

        sage: G = QuaternionGroup()
        sage: mats = real_isotypic_projection_matrices(G, 8)
        sage: sum(mats).is_one()
        True
    """
    return [isotypic_projection_matrix(G, d, chi, deg) for chi,deg in zip(*real_characters(G))]




r"""
Plane trees.

Generation of plane trees
-------------------------

This file contains generators for level sequences of

 * rooted (free) trees
 * (free) trees
 * rooted plane trees
 * plane trees

REFERENCES:

  * Beyer-Hedetniemi "Constant time generation of rooted trees", 1980
  Generation of rooted trees (a root but no cyclic order at vertices). They use
  a successor function on level sequences.
  (good explanation of the algorithm)
  A regular tree is a tree such that for all vertices, its sequence of subtrees
  is in decreasing order.
  How do PREV and SAVE work ?

  * Wright-Richmond-Odlyzko-McKay "Constant time generation of free trees", 1986
  TODO: look at

.. [Nak02]   Nakano "Efficient generation of plane trees", 2002
  For him, rooted means left-right ordering of children of each node.
  Enumerating rooted plane trees with at most `n` vertices is elementary (a
  depth-first search in the tree of trees).
"""
#*****************************************************************************
#       Copyright (C) 2016-2019 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

from __future__ import print_function

from sage.rings.integer import Integer

###################################
# Rooted trees (Beyer-Hedetniemi) #
###################################

def _TMP_successor(l):
    r"""
    given a level sequence generate the next (Beyer-Hedetniemi)
    """
    # find p the last index in l which is not 1
    p = len(l)-1
    while p > 0 and l[p] == 1:
        p -= 1
    if p == 0:
        return False
    # find q the parent of p
    q = p-1
    while q > 0 and l[q] != l[p]-1:
        q -= 1
    # build the new level sequence
    # l[:p] does not change and then copy as many times as you can the level
    # sequence l[q:p].
    diff = p-q
    ll = l[:p] + l[q:p]*((len(l)-p)/diff)
    ll.extend(l[q:q+len(l)-len(ll)])

    return ll

def _TMP_rooted_trees(n):
    r"""
    Return the list of rooted trees using the preceding function
    """
    l = range(n)
    res = [l]
    ll = _TMP_successor(res[-1])
    while ll:
        res.append(ll)
        ll = _TMP_successor(res[-1])
    return res

def rooted_tree_iterator(n,verbose=False):
    r"""
    Iterator through regular level sequences of rooted trees.
    (only works for n >= 3)

    EXAMPLES::

        sage: from surface_dynamics.misc.plane_tree import rooted_tree_iterator
        sage: for t in rooted_tree_iterator(4): print(t)
        [0, 1, 2, 3]
        [0, 1, 2, 2]
        [0, 1, 2, 1]
        [0, 1, 1, 1]
        sage: for t in rooted_tree_iterator(5): print(t)
        [0, 1, 2, 3, 4]
        [0, 1, 2, 3, 3]
        [0, 1, 2, 3, 2]
        [0, 1, 2, 3, 1]
        [0, 1, 2, 2, 2]
        [0, 1, 2, 2, 1]
        [0, 1, 2, 1, 2]
        [0, 1, 2, 1, 1]
        [0, 1, 1, 1, 1]

        sage: for t in rooted_tree_iterator(5,verbose=True): pass
          p =    4
          prev = [0, 1, 2, 3, 4]
          save = [0, 0, 0, 0, 0]
        [0, 1, 2, 3, 4]
          p =    4
          prev = [0, 1, 2, 3, 4]
          save = [0, 0, 0, 0, 0]
        [0, 1, 2, 3, 3]
          p =    4
          prev = [0, 1, 2, 3, 4]
          save = [0, 0, 0, 0, 0]
        [0, 1, 2, 3, 2]
          p =    3
          prev = [0, 1, 2, 0, 4]
          save = [0, 0, 0, 0, 0]
        [0, 1, 2, 3, 1]
          p =    4
          prev = [0, 1, 3, 0, 4]
          save = [0, 0, 0, 2, 0]
        [0, 1, 2, 2, 2]
          p =    3
          prev = [0, 1, 2, 0, 4]
          save = [0, 0, 0, 2, 0]
        [0, 1, 2, 2, 1]
          p =    4
          prev = [0, 3, 2, 0, 4]
          save = [0, 0, 0, 1, 0]
        [0, 1, 2, 1, 2]
          p =    2
          prev = [0, 1, 0, 0, 4]
          save = [0, 0, 0, 1, 0]
        [0, 1, 2, 1, 1]
          p =    0
          prev = [0, 0, 0, 0, 4]
          save = [0, 0, 0, 1, 0]
        [0, 1, 1, 1, 1]
    """
    assert n >= 3

    l = list(range(n))
    prev = list(range(n))  # function: level -> ?
    save = [0]*n
    p = n-1

    if verbose:
        print("  p =    %s" % p)
        print("  prev = %s" % prev)
        print("  save = %s" % save)
        print(l)
    yield l

    while p > 0:
        l[p] = l[p] - 1
        if p < n and (l[p] != 1 or l[p-1] != 1):
            diff = p - prev[l[p]] # = p-q
            while p < n-1:
                save[p] = prev[l[p]]
                prev[l[p]] = p
                p += 1
                l[p] = l[p-diff]
        while l[p] == 1:
            p -= 1
            prev[l[p]] = save[p]

        if verbose:
            print("  p =    %s" % p)
            print("  prev = %s" % prev)
            print("  save = %s" % save)
            print(l)
        yield l

###############################
# Rooted plane trees (Nakano) #
###############################

def rooted_plane_tree_iterator(nmin,nmax=None, verbose=False):
    r"""
    Iterator through rooted plane trees with at least ``nmin`` and at most
    ``nmax`` vertices. If ``nmax`` is not furnished then it is automatically set
    to ``nmin``.

    The output are list which corresponds to level sequences. The algorithm,
    which corresponds to a depth first search in a tree of trees is due to
    Nakano [Nak02]_.

    EXAMPLES::

        sage: from surface_dynamics.misc.plane_tree import rooted_plane_tree_iterator

        sage: for t in rooted_plane_tree_iterator(4): print(t)
        [0, 1, 2, 3]
        [0, 1, 2, 2]
        [0, 1, 2, 1]
        [0, 1, 1, 2]
        [0, 1, 1, 1]
        sage: for t in rooted_plane_tree_iterator(1,3): print(t)
        [0]
        [0, 1]
        [0, 1, 2]
        [0, 1, 1]
    """
    nmin = int(nmin)
    if nmax is None:
        nmax = nmin
    else:
        nmax = int(nmax)
        if nmin > nmax:
            raise ValueError("nmin (=%d) should be lesser than nmax (=%d)"%(nmin,nmax))
    l = [0]
    b = [] # current branch
    while True:
        if len(l) >= nmin:
            yield l
        if len(l) < nmax: # go down
            if verbose: print("go down")
            if verbose: print(" append: %d"%(l[-1]+1))
            l.append(l[-1]+1)
            b.append(l[-1])
        else: # go up
            if verbose: print("go up")
            while b and b[-1] == 1:
                if verbose: print(" pop")
                b.pop(-1)
                l.pop(-1)
            if b == []:
                break
            if verbose: print(" modify last to become %d"%b[-1])
            b[-1] -= 1
            l[-1] = b[-1]

def half_turn(t,i=None):
    r"""
    Half turn of canonical form of a tree t which has an odd diameter.

    INPUT:

    - ``s`` - position of 1 in t

    - ``i`` - the guy of depth m-1 as an index in s

    EXAMPLES::

        sage: from surface_dynamics.misc.plane_tree import half_turn

        sage: half_turn([0,1,2,2,2,1])
        [0, 1, 2, 1, 1, 1]
        sage: half_turn([0,1,2,1,1,1])
        [0, 1, 2, 2, 2, 1]

        sage: half_turn([0,1,2,3,2,1,1,2,1])
        [0, 1, 2, 2, 3, 2, 1, 2, 1]
        sage: half_turn([0,1,2,2,3,2,1,2,1])
        [0, 1, 2, 3, 2, 1, 1, 2, 1]
    """
    if i is None:
        i = 2
        while i < len(t) and t[i] != 1:
            i += 1
    return [0,1] + [j+1 for j in t[i:]] + [j-1 for j in t[2:i]]

def cmp_halves(t,i=None):
    r"""
    Compare the subtrees [2:i] and [i:].

    Used in the case of a central edge in the tree.
    """
    t1 = [j-1 for j in t[2:i]]
    t2 = t[i:]
    return (t1 > t2) - (t1 < t2)

def cmp_subtree(t1,d1,t2,d2):
    test = (d1 > d2) - (d1 < d2)
    if test: return test

    return (t1 > t2) - (t1 < t2)

def is_lyndon(t,s=None,d=None,verbose=False):
    r"""
    Returns
    Lyndon -> True
    periodic -> period
    otherwise -> False

    ALGORITHM

    Lothaire, Combinatorics on words

    INPUT:

    - ``t`` - the tree

    - ``s`` - the positions of 1 in t (limit of subtrees)

    - ``d`` - the depths of subtrees

    EXAMPLES::

        sage: from surface_dynamics.misc.plane_tree import is_lyndon

        sage: is_lyndon([1,2,3,3,1,2])
        True
        sage: is_lyndon([1,2,3,2,3,4,1,2,1,2,3,4])
        False

        sage: is_lyndon([1,1,2,1,2])
        False
        sage: is_lyndon([1,2,1,1,2])
        False
        sage: is_lyndon([1,2,1,2,1])
        True

        sage: is_lyndon([1,1,2,3,2,1,2,3,2,1,2,3,2])
        False
        sage: is_lyndon([1,2,3,2,1,1,2,3,2,1,2,3,2])
        False
        sage: is_lyndon([1,2,3,2,1,2,3,2,1,1,2,3,2])
        False
        sage: is_lyndon([1,2,3,2,1,2,3,2,1,2,3,2,1])
        True

        sage: is_lyndon([1,2,1,2])
        1

        sage: is_lyndon([1,2,3,1,2,1,2,3,1,2]) # lyndon periodic
        2
        sage: is_lyndon([1,2,1,2,3,1,2,1,2,3]) # not lyndon periodic
        False
    """
    if s is None:
        s = [i for i in range(len(t)) if t[i] == 1]
    n = len(s)
    if n <= 1:
        return True
    w = [t[s[i]+1:s[i+1]] for i in range(len(s)-1)]
    w.append(t[s[-1]+1:])
    if d is None:
        d = [max([1]+ww) for ww in w]
    if verbose:
        print("w=",w)
    i,j=0,1
    while j < n:
        if verbose:
            print(" i,j =",i,j)
            print(" w[i] =",w[i])
            print(" w[j] =",w[j])
        c = cmp_subtree(w[i],d[i],w[j],d[j])
        if c == 0:
            i += 1
            j += 1
        elif c > 0:
            i = 0
            j += 1
        else:
            return False
    # i = 0: Lyndon, j-i = i: periodic ?
    if verbose:
        print(" out", i,j)
    if i == 0:
        return True
    elif n%(j-i) == 0:
        return j-i
    return False

def unrooted_plane_tree_iterator(nmin,nmax=None,verbose=False, check=False):
    r"""
    Iterator through plane trees with at most nmax vertices. A representative of
    each equivalence class of rooted tree is returned. The order is decreasing
    among level sequences.

    One can choose a canonical representative by choosing either the middle
    vertex as a root (and the maximal cyclic order among subtrees) or the middle
    edge as the left-most edge (and we have to check between two).

    Sloane's A002995
    1, 1, 1, 1, 2, 3, 6, 14, 34, 95, 280, 854, 2694, 8714, 28640, 95640, 323396,
    1105335, 3813798, 13269146

    TODO:

    one more restriction for go down: if the max depth is attained only at first
    position and there is not enough available vertices to go deeply enough.

    EXAMPLES::

        sage: from surface_dynamics.misc.plane_tree import unrooted_plane_tree_iterator
        sage: for t in unrooted_plane_tree_iterator(4): print(t)
        [0, 1, 2, 1]
        [0, 1, 1, 1]
        sage: for t in unrooted_plane_tree_iterator(5): print(t)
        [0, 1, 2, 2, 1]
        [0, 1, 2, 1, 2]
        [0, 1, 1, 1, 1]
        sage: for t in unrooted_plane_tree_iterator(6): print(t)
        [0, 1, 2, 3, 1, 2]
        [0, 1, 2, 2, 2, 1]
        [0, 1, 2, 2, 1, 2]
        [0, 1, 2, 2, 1, 1]
        [0, 1, 2, 1, 2, 1]
        [0, 1, 1, 1, 1, 1]

        sage: for t in unrooted_plane_tree_iterator(1,5): print(t)
        [0]
        [0, 1]
        [0, 1, 2, 2, 1]
        [0, 1, 2, 1]
        [0, 1, 2, 1, 2]
        [0, 1, 1]
        [0, 1, 1, 1]
        [0, 1, 1, 1, 1]

        sage: [sum(1 for _ in unrooted_plane_tree_iterator(i)) for i in range(1,13)]
        [1, 1, 1, 2, 3, 6, 14, 34, 95, 280, 854, 2694]
    """
    nmin = int(nmin)
    if nmax is None:
        nmax = nmin
    else:
        nmax = int(nmax)
        if nmin > nmax:
            raise ValueError("nmin (=%d) should be lesser than nmax (=%d)"%(nmin,nmax))
    nnmax = nmax//2

    #TODO: start at the right place and remove trivial cases in recursion
    t = [0]    # current tree
    b = []     # current branch in the tree of trees
    s = []     # positions of 1 in t (or limit of subtrees)
    d = []     # depths of subtrees of t
    d_occ = [] #TODO: nb of occurrences in d
    while True:
        if verbose:
            print("  t = %s" % t)
            print("  d = %s" % d)
            print("  s = %s" %s)
        # verifications
        if check:
            assert len(d) == len(s)
            assert all(t[i] == 1 for i in s)
            assert (len(d) == 0 and t == [0]) or (max(d) == max(t))
        # do we need to output t
        if len(t) >= nmin:
            if t == [0]: #trivial case
                yield t
            else:
                #TODO: remove the loop and use d_occ
                m = d[0]       # depth of first subtree
                m_nb = 1       # nb of subtrees with depth m
                has_mm = False # is there a subtree with depth m-1
                for i in range(1,len(d)):
                    if d[i] > m: # the deepest branch is not the first one
                        break
                    elif d[i] == m:
                        m_nb += 1
                    elif d[i] == m-1:
                        has_mm = True
                else: # we did not break the loop
                    if ((m == 1) or
                        (m_nb == 1 and has_mm and cmp_halves(t,s[1]) >= 0) or
                        (m_nb > 1 and is_lyndon(t,s,d))):
                        if verbose: print("YIELD")
                        yield t
                    elif verbose:
                        print("NOT YIELD")
        # do we go down
        #TODO: remove len(s) <= 1 (corresponds to the leftmost branch in the
        #      tree of trees)
        if (len(t) < nmax and
            ((len(s) <= 1 and t[-1] <= nnmax) or
             (len(s) > 1 and t[-1] <= d[0]))):
            if verbose: print("go down")
            if verbose: print(" append: %d"%(t[-1]+1))
            #TODO: append the right thing which is not necessarily t[-1]+1
            t.append(t[-1]+1)
            b.append(t[-1])
            if t[-1] == 1: # trivial case (t was [0] before)
                if check:
                    assert len(s) == 0
                    assert len(d) == 0
                s.append(len(t)-1)
                d.append(1)
            elif t[-1] > d[-1]:
                d[-1] = t[-1]
        # or go up
        else:
            if verbose: print("go up")
            while b and b[-1] == 1:
                if verbose: print(" pop")
                b.pop(-1)
                t.pop(-1)
                if check:
                    assert d[-1] == 1
                    assert s[-1] == len(t)
                d.pop(-1)
                s.pop(-1)
            if b == []:
                break
            if verbose:
                print(" modify last to become %d"%(b[-1]-1))
            b[-1] -= 1
            t[-1] = b[-1]
            if d[-1] == t[-1]+1: # update the depth of the rightmost branch or
                                 # the one before if t[-1] == 1
                i = 1
                d[-1] = 1
                for i in range(s[-1]+1,len(t)-1):
                    if t[i] > d[-1]:
                        d[-1] = t[i]
                if verbose: print(" updated length of rightmost which is now %d"%d[-1])
            if t[-1] == 1: # new branch
                d.append(1)
                s.append(len(t)-1)

##########################################################################
# Modified version of Nakano's algorithm to generate separatrix diagrams #
##########################################################################

def _TMP_admissible_plane_tree_iterator(a, verbose=False):
    r"""
    Iterator through plane trees with at most nmax vertices.

    INPUT:

    - ``a`` - the angle at the vertices divided by pi.

    OUTPUT:

    - ``(t,n,l)`` - ``t`` is a tree, ``n`` its number of nodes and ``l`` its
      number of leaves.

    ALGORITHM:

    variables in the algorithm

    t : the current tree
    b : the current branch explored in the tree of trees
    n : number of nodes of t except the root
    l : number of leaves of t

    the tree is valid if 2n >= a and 2n-l <= a
    """
    t = [0] # the tree
    b = []  # current branch
    n = 0   # number of nodes (=len(t)-1)
    l = 1   # number of leaves
    while True:
        if verbose:
            print(t)
            print("n = %2d,   l = %2d" %(n,l))
        if (2*n >= a) and (2*n-l <= a):
            if verbose: print("yield")
            t.append(0)
            yield t, n, l
            t.pop(-1)

        if n <= a and 2*n-l < a: # could add more nodes -> go down
            if verbose:
                print("go down")
                print("  append: %d"%(t[-1]+1))
            t.append(t[-1]+1)
            b.append(t[-1])
            n += 1
            # l is unchanged

        else: # we are too far -> go up
            if verbose: print("go up")
            # update l and n, then pop the rightmost node
            while b and b[-1] == 1:
                if verbose: print(" pop")
                b.pop(-1)
                t.pop(-1)
                n -= 1
                l -= 1
            if b == []:
                break
            if verbose:
                print(" modify last")
            if len(t) > 1 and t[-2] == t[-1] - 1:
                l += 1
            b[-1] -= 1
            t[-1] = b[-1]

        if verbose: print()

def admissible_plane_tree_iterator(a,verbose=False):
    r"""
    Plane tree for separatrix diagram of hyperelliptic strata. Return triple
    (t,n,l) where t is a tree, n its number of nodes and l its number of leaves.

    A plane tree with n nodes and l leaves is *a-admissible* if it satisfies
    `n <= a` and `2n-l < a`

    EXAMPLES::

        sage: from surface_dynamics.misc.plane_tree import admissible_plane_tree_iterator
        sage: for t,n,l in admissible_plane_tree_iterator(5): print("%s\n%s %s" %(t,n,l))
        [0, 1, 2, 2, 1, 0]
        4 3
        [0, 1, 2, 1, 0]
        3 2
        [0, 1, 1, 1, 0]
        3 3
        [0, 1, 1, 1, 1, 0]
        4 4
        [0, 1, 1, 1, 1, 1, 0]
        5 5
    """
    aa = (a+1)//2

    #TODO: start at the right place and remove trivial cases in recursion
    t = [0]    # current tree
    b = []     # current branch in the tree of trees
    s = []     # positions of 1 in t (or limit of subtrees)
    d = []     # depths of subtrees of t
    n = 0      # number of nodes (=len(t)-1)
    l = 1      # number of leaves
    d_occ = [] #TODO: nb of occurrences in d
    while True:
        if verbose:
            print("  t = %s" % t)
            print("  d = %s" % d)
            print("  s = %s" % s)
        # do we need to output t
        if (2*n >= a) and (2*n-l <= a):
            #TODO: remove the loop and use d_occ
            m = d[0]       # depth of first subtree
            m_nb = 1       # nb of subtrees with depth m
            has_mm = False # is there a subtree with depth m-1
            for i in range(1,len(d)):
                if d[i] > m: # the deepest branch is not the first one
                    break
                elif d[i] == m:
                    m_nb += 1
                elif d[i] == m-1:
                    has_mm = True
            else: # we did not break the loop
                if ((m == 1) or
                    (m_nb == 1 and has_mm and cmp_halves(t,s[1]) >= 0) or
                    (m_nb > 1 and is_lyndon(t,s,d))):
                    if verbose: print("YIELD")
                    t.append(0)
                    yield t,n,l
                    t.pop(-1)
                elif verbose:
                    print("NOT YIELD")
        # do we go down
        #TODO: remove len(s) <= 1 (corresponds to the leftmost branch in the
        #      tree of trees)
        if ((n <= a and 2*n-l < a) and
           ((len(s) <= 1 and t[-1] <= aa) or (len(s) > 1 and t[-1] <= d[0]))):
            if verbose: print("go down")
            if verbose: print(" append: %d"%(t[-1]+1))
            #TODO: append the right thing which is not necessarily t[-1]+1
            t.append(t[-1]+1)
            b.append(t[-1])
            n += 1
            if t[-1] == 1: # trivial case (t was [0] before)
                s.append(len(t)-1)
                d.append(1)
            elif t[-1] > d[-1]:
                d[-1] = t[-1]
        # or go up
        else:
            if verbose: print("go up")
            while b and b[-1] == 1:
                if verbose: print(" pop")
                b.pop(-1)
                t.pop(-1)
                d.pop(-1)
                s.pop(-1)
                n -= 1
                l -= 1
            if b == []:
                break
            if verbose:
                print(" modify last to become %d"%(b[-1]-1))
            if len(t) > 1 and t[-2] == t[-1] - 1:
                l += 1
            b[-1] -= 1
            t[-1] = b[-1]
            if d[-1] == t[-1]+1: # update the depth of the rightmost branch or
                                 # the one before if t[-1] == 1
                i = 1
                d[-1] = 1
                for i in range(s[-1]+1,len(t)-1):
                    if t[i] > d[-1]:
                        d[-1] = t[i]
                if verbose: print(" updated length of rightmost which is now %d"%d[-1])
            if t[-1] == 1: # new branch
                d.append(1)
                s.append(len(t)-1)




r"""
Kontsevich topological recursion that computes intersection of psi classes.

EXAMPLES::

    sage: from surface_dynamics import KontsevichTR
    sage: K = KontsevichTR()

    sage: for p in [(4,),(5,0),(4,1),(3,2)]:
    ....:     n = len(p)
    ....:     assert sum(p) == n + 3
    ....:     c = K.F(2, n, p) / prod((2*i + 1).multifactorial(2) for i in p)
    ....:     print(p, c)
    (4,) 1/1152
    (5, 0) 1/1152
    (4, 1) 1/384
    (3, 2) 29/5760
"""
#*****************************************************************************
#       Copyright (C) 2020 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************
from sage.misc.misc_c import prod
from sage.structure.unique_representation import UniqueRepresentation
from sage.rings.all import ZZ, QQ
from .topological_recursion import TopologicalRecursion

ZZ_1 = ZZ.one()

def psi_correlator(*args):
    r"""
    Return the integral of psi classes

    EXAMPLES:

    Examples in genus 0::

        sage: from surface_dynamics.topological_recursion.kontsevich import psi_correlator
        sage: psi_correlator(0,0,0)
        1
        sage: psi_correlator(1,0,0,0)
        1
        sage: psi_correlator(2,0,0,0,0)
        1
        sage: psi_correlator(1,1,0,0,0)
        2

    Examples in genus 1::

        sage: psi_correlator(1)
        1/24
        sage: psi_correlator(2, 0)
        1/24
        sage: psi_correlator(1, 1)
        1/24
        sage: psi_correlator(3, 0, 0)
        1/24
        sage: psi_correlator(2, 1, 0)
        1/12
        sage: psi_correlator(1, 1, 1)
        1/12

    genus 2::

        sage: psi_correlator(7)
        1/82944
        sage: psi_correlator(7, 1)
        5/82944
        sage: psi_correlator(6, 2)
        77/414720
        sage: psi_correlator(5, 3)
        503/1451520
        sage: psi_correlator(4, 4)
        607/1451520
    """
    assert all(x in ZZ and x >= 0 for x in args)
    n = len(args)
    s = sum(args)
    if (s - n) % 3:
        raise ValueError("the composition should sum up to 3*g - 3 + n")
    g = (s -n)//3 + 1
    return KontsevichTR().F(g, n, args) / prod((2*i + 1).multifactorial(2) for i in args)

class KontsevichTR(UniqueRepresentation, TopologicalRecursion):
    r"""
    Topological recursion for intersection of psi classes (Witten's conjecture)

    `\int_{Mgn} exp(L_i^2/2 psi_i)`

    EXAMPLES::

        sage: from surface_dynamics import KontsevichTR
        sage: K = KontsevichTR()

        sage: [K.F(2, 1, (i,)) for i in range(5)]
        [0, 0, 0, 0, 105/128]

        sage: K.polynomial(0,3)
        1
        sage: K.polynomial(0,4)
        3*b0 + 3*b1 + 3*b2 + 3*b3
        sage: K.polynomial(0,5)
        15*b0^2 + 18*b0*b1 + ...

        sage: K.polynomial(1,1)
        1/8*b0
        sage: K.polynomial(1,2)
        5/8*b0^2 + 3/8*b0*b1 + 5/8*b1^2

        sage: K.write(0, 4)
        [1, 0, 0, 0] 1/2
        sage: K.write(0, 5)
        [2, 0, 0, 0, 0] 1/8
        [1, 1, 0, 0, 0] 1/2
        sage: K.write(1, 2)
        [2, 0] 1/192
        [1, 1] 1/96
        sage: K.write(2, 1)
        [4] 1/442368
    """
    def __init__(self):
        TopologicalRecursion.__init__(self)

    def A(self, i, j, k):
        r"""
        A-data for Kontsevich recursion
        """
        if i == 0 and j == 0 and k == 0:
            return ZZ_1
        else:
            raise ValueError

    def B(self, g, n, i, j):
        r"""
        B-data for Kontsevich topological recursion
        """
        # before we had s=I[0]+I[m]=i+j
        # i + j = k + 1
        k = i + j - 1
        if k >= 0:
            yield (k, ZZ(2*j + 1))

    def C(self, i, jmax, kmax, smax):
        r"""
        C-data for Konstevich topological recursion
        """
        if i < 2 or jmax > i-2 or kmax > i-2:
            # this is a waste of time...
            return

        # iterate through indices (j,k) with j + k = i-2
        jmin = max(0, i - 2 - kmax)
        jmax = min(i - 2, jmax)
        for j in range(jmin, jmax+1):
            yield (j, i-j-2, ZZ_1)

    def D(self, i):
        r"""
        D-data for Kontsevich topological recursion
        """
        return QQ(((i == 1), 8))

